% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file cse.pl
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% common sub-expressions elimination
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% commutativity normalisation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
cse_commutativity(X, X) :-
	simple(X),
	!.
cse_commutativity(min(A1, B1), min(A3, B3)) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	sort2(A2, B2, A3, B3).
cse_commutativity(max(A1, B1), max(A3, B3)) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	sort2(A2, B2, A3, B3).
cse_commutativity(+(A1, B1), +(A3, B3)) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	sort2(A2, B2, A3, B3).
cse_commutativity(*(A1, B1), *(A3, B3)) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	sort2(A2, B2, A3, B3).
cse_commutativity(and(A1, B1), Norm) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	(
	  A2 == B2
	->
	  Norm = A2
	;
	  Norm = and(A2, B2)
	).
cse_commutativity(or(A1, B1), Norm) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	(
	  A2 == B2
	->
	  Norm = A2
	;
	  Norm = or(A2, B2)
	).
cse_commutativity(equiv(A1, B1), Norm) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	(
	  A2 == B2
	->
	  Norm = A2
	;
	  Norm = equiv(A2, B2)
	).

/*
disabled : Don't want to reorder formulae code...
-- Julien
cse_commutativity( and(A1,labeling(L)), and(A2,labeling(L))) :-
	!,
	cse_commutativity(A1, A2).
cse_commutativity( and(A1,B1), Norm) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	(   A2==B2 -> Norm = A2
	;   Norm = and(A3,B3),
	    sort2(A2, B2, A3, B3)
	).
cse_commutativity( or(A1,B1), Norm) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	(   A2==B2 -> Norm = A2
	;   Norm = or(A3,B3),
	    sort2(A2, B2, A3, B3)
	).
cse_commutativity( or(A1,B1), or(A3,B3)) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	sort2(A2, B2, A3, B3).
cse_commutativity( equiv(A1,B1), Norm) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	(   A2==B2 -> Norm = true
	;   Norm = equiv(A3,B3),
	    sort2(A2, B2, A3, B3)
	).
cse_commutativity( equiv(A1,B1), equiv(A3,B3)) :-
	!,
	cse_commutativity(A1, A2),
	cse_commutativity(B1, B2),
	sort2(A2, B2, A3, B3).
*/

cse_commutativity(X, Y) :-
	functor(X, F, A),
	functor(Y, F, A),
	cse_commutativity_args(0, A, X, Y).

cse_commutativity_args(K, K, _, _) :-
	!.
cse_commutativity_args(I, K, X, Y) :-
	J is I+1,
	arg(J, X, A),
	arg(J, Y, B),
	cse_commutativity(A, B),
	cse_commutativity_args(J, K, X, Y).
	
sort2(A, B, C, D) :-
	(
	  A @=< B
	->
	  C = A,
	  D = B
	;
	  C = B,
	  D = A
	).

% common sub expression elimination
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
cse(Tree1, Tree3) :-
	empty_avl(AVL1),
	State1 = (AVL1, 0),
	cse_scan(Tree1, State1, (AVL2, _), 0, Trace, []),
	cse_replace(Tree1, Tree2, AVL2, 0),	
	cse_and(Trace, AVL2, Tree2, Tree3).

cse_scan(X, State, State, _) -->
	{
	 simple(X)
	},
	!.
cse_scan(X, State, State, _) -->
	{
	 is_list(X)
	},
	!.
cse_scan(X, (AVL1, G1), (AVL2, G2), Level) -->
	{
	 avl_fetch(X, AVL1, (Y, Top))
	},
	!,
	{
	 cse_mark_top(Level, Top),
	 AVL2 = AVL1
	},
	(
	  {
	   nonvar(Y)
	  }
	->
	  {	   
	   G2 = G1
	  }
	;
	  [X],
	  {
	   G2 is G1 + 1,
	   number_codes(G2, Trees),
	   atom_codes(Y, [0'G | Trees])
	  }
	).
cse_scan(and(X, Y), State0, State, 0) -->
	!,
	cse_scan(X, State0, State1, 0),
	cse_scan(Y, State1, State, 0).
cse_scan(X, (AVL1, G1), State, Level) -->
	{
	 avl_store(X, AVL1, (_, Top), AVL2),
	 cse_mark_top(Level, Top),
	 functor(X, _, Ar),
	 Level1 is Level + 1
	},
	cse_scan_args(0, Ar, X, (AVL2, G1), State, Level1).

cse_scan_args(K, K, _, State, State, _) -->
	!.
cse_scan_args(I, K, X, State0, State, L) -->
	{
	 J is I + 1,
	 arg(J, X, Y)
	},
	cse_scan(Y, State0, State1, L),
	cse_scan_args(J, K, X, State1, State, L).

cse_mark_top(0, top) :-
	!.
cse_mark_top(_, _).

cse_replace(X, X, _, _) :-
	simple(X),
	!.
cse_replace(X, X, _, _) :-
	is_list(X),
	!.
cse_replace(and(X1, Y1), and(X2, Y2), AVL, 0) :-
	!,
	cse_replace(X1, X2, AVL, 0),
	cse_replace(Y1, Y2, AVL, 0).
cse_replace(X, Z, AVL, Level) :-
	avl_fetch(X, AVL, (Y, Top)),
	(
	  (
	    nonvar(Y),
	    Top == top
	  )
	->
	  Z = true
	;
	  nonvar(Y)
	->	  
	  Z = Y
	;
	  functor(X, F, Ar),
	  functor(Z, F, Ar),
	  Level1 is Level + 1,
	  cse_replace_args(0, Ar, X, Z, AVL, Level1)
	).

cse_replace_args(K, K, _, _, _, _) :-
	!.
cse_replace_args(I, K, X, Y, AVL, L) :-
	J is I + 1,
	arg(J, X, A),
	arg(J, Y, B),
	cse_replace(A, B, AVL, L),
	cse_replace_args(J, K, X, Y, AVL, L).

cse_and([], _, Tree, Tree).
cse_and([X | Xs], AVL, Tree1, and(Def, Tree2)) :-
	avl_fetch(X, AVL, (Atom,Top)),
	functor(X, F, Ar),
	functor(Y, F, Ar),
	cse_def(F, Atom, Top, Y, Def),
	cse_replace_args(0, Ar, X, Y, AVL, 0),
	cse_and(Xs, AVL, Tree1, Tree2).

cse_def(_, _, Top, Def, Def) :-
	Top == top,
	!.
cse_def(or,    A, _, Y, equiv(A, Y)) :-
	!.
cse_def(and,   A, _, Y, equiv(A, Y)) :-
	!.
cse_def(equiv, A, _, Y, equiv(A, Y)) :-
	!.
cse_def(F,     A, _, Y, equiv(A, Y)) :-
	compop_neg(F, _),
	!.
cse_def(_,     A, _, Y, A = Y).
