% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file lang.pl
%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% connectives, functions & relations
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% statement
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(900, fx, ?).

% such that 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(800, xfx, if).

% logics
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(780, xfx, implies).
:- op(780, xfx, equiv).
:- op(775, xfy, or).
:- op(770, xfy, and).
:- op(710,  fy, not).

% comparison
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(700, xfx, #).

% domain
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(700, xfx, in).

% sets
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(700, xfx, =-=).
:- op(600, yfx, \-\).
:- op(500, yfx, \-/).
:- op(400, yfx, /-\).

% range
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
:- op(550, xfx, ..).
:- op(550, xfy, ::).

:- op(100, xfy, :).

% builtin operators
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(Key, Val) :-
	lang(Key, Val, []).

lang(comp_ops) --> [(<)/2, (=<)/2, (=)/2, (>=)/2, (>)/2, (#)/2].

lang(arith_ops) --> [(+)/2, (-)/2, (*)/2, (/)/2, max/2, min/2, abs/1, log/2, exp/2].

lang(logic_ops) --> [not/1, and/2, or/2, implies/2, equiv/2].

lang(conjdisj_ops) --> [and/2, or/2].

lang(list_ops) --> [(.)/2, (.)/0, cons/0, cons/2]. 

lang(set_ops) --> [/-\, \-/, \-\, =-=, card].

lang(builtins) --> 
	lang(comp_ops),
	lang(arith_ops),
	lang(logic_ops),
	lang(set_ops),
	lang(list_ops).

% variable binding predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(bind_preds) --> [foldr, foldl, map, aggregate, forall, exists, let, if].

% logical quantification predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(quant_preds) --> [forall/3, exists/3, let/3].

% higher-order functions
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(hof_preds) --> [foldr/5, foldl/5].

% variable domain predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(domain_preds) --> [in/2, domain/1, domain/3].

% variable domain predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(range_preds) --> [in/2, domain/1, domain/3].

% labeling-search-optim predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(label_preds) --> [labeling/1].

lang(stree_preds) --> [search/1].

lang(optim_preds) --> [minimize/2, maximize/2].

lang(search_preds) -->
	lang(label_preds),
	lang(stree_preds),
	lang(optim_preds).

% heuristics predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% var-val 
%%%%%%%%%%%%%%%%%%%%
lang(varorder_preds) --> [variable_ordering/1].
lang(varordercrit_preds) --> [is/1, any/1, least/1, greatest/1].

lang(valorder_preds) --> [value_ordering/1].
lang(valordercrit_preds) --> [up/1, down/1, bisect/1, step/1, enum/1].

lang(vorder_preds) -->
	lang(varorder_preds),
	lang(valorder_preds).

lang(vordercrit_preds) -->
	lang(varordercrit_preds),
	lang(valordercrit_preds).

lang(dynvarordercrit_preds) --> [var_order_criterion/_].
lang(dynvalordercrit_preds) --> [val_order_criterion/_].
lang(dynvordercrit_preds) -->
	lang(dynvarordercrit_preds),
	lang(dynvalordercrit_preds).	

% conj-disj
%%%%%%%%%%%%%%%%%%%%
lang(conjorder_preds) --> [conjunct_ordering/1].

lang(disjorder_preds) --> [disjunct_ordering/1].

lang(folorder_preds) -->
	lang(conjorder_preds),
	lang(disjorder_preds).

lang(folordercrit_preds) --> [least/1, greatest/1].

lang(order_preds) -->
	lang(vorder_preds),
	lang(vordercrit_preds),
	lang(folorder_preds),
	lang(folordercrit_preds).

% gui predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(gui_varname_preds) --> [gui_name/2].

lang(gui_control_preds) --> [gui_control/3].

lang(gui_preds) -->
	lang(gui_varname_preds),
	lang(gui_control_preds).

% global constraint predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(geosr_preds) --> [geosr_non_overlapping/3].

lang(glbcstr_preds) --> [all_different/1,
			 disjoint2/3,
			 cumulative/4,
			 geost_constraint/3,
			 geost_constraint/4,
			 lexicographic/1,
			 lexicographic_strict/1].

% list related predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(list_preds) --> [length/1, nth/2, pos/2].

lang(pl_preds) --> [','/2, ';'/2, '.'/2].

lang(dbg_preds) --> [rcp_write/1].

% inlined predicates
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
lang(inlined_preds) -->	
	lang(geosr_preds),
	lang(glbcstr_preds),
	lang(gui_preds),
	lang(domain_preds),
	lang(order_preds),
	lang(search_preds),
	lang(pl_preds),
	lang(dbg_preds).
