% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file ocg.cj.pl
%
:- dynamic first_labeling/1,
           cj_solver_cstr/1,
	   cj_solver_goal/1.

% CLPFD Choco Java operators dictionary 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
dict_cj(<, lt).
dict_cj(=<, leq).
dict_cj(=, eq).
dict_cj(#, neq).
dict_cj(>=, geq).
dict_cj(>, gt).
dict_cj(true,  true).
dict_cj(false, false).
dict_cj(and, and).
dict_cj(or, or).
dict_cj(not, not).
dict_cj(equiv, ifOnlyIf).
dict_cj(implies, implies).
dict_cj(+, plus).
dict_cj(-, minus).
dict_cj(*, mult).
dict_cj(/, div).
dict_cj(min, min).
dict_cj(max, max).
dict_cj(abs, abs).

% atom 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(Term, STerm, []) :-	
	atom(Term),
	atom_codes(Term, [0'$ | T]),
	!,
	atom_codes(STerm, T).

% variable 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(var(X, I), STerm, []) :-
	cg_mode(dynamic),
	!,		
	serialize_term(X, SX),
	to_upper_case(SX, SXUpp),
	serialize_int_var_sufx(I, SI),
	atomize([SXUpp, SI], '_', STerm).

% free variable (FD var.)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(free_var(X, _, I), Var,
		 [cj_var_decl(cj_class(integerVariable), Var) =
		 rcp_var(cj_string(from(Y, I)))] ) :- 
	!,
	delete_uids(X, Y),
	serialize_term(Y, SX),
	serialize_int_var_sufx(I, SI),
	atomize([SX, SI], '_', Var).

% % dynamic tag
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate_cj_code(dynamic(Term), STerm, []) :-
% 	!,
% 	set_cg_mode(dynamic),
% 	generate_cj_code(Term, STerm1, SAdd1),
% 	(	  
% 	  is_var_name(STerm1)
% 	->
% 	  append([rcp_fol_call(STerm1)], SAdd1, SAdd),
% 	  list2seq(SAdd, STerm)
% 	;	  
% 	  list2seq(SAdd1, STerm0),
% 	  STerm = (STerm0, STerm1)
% 	).

% search
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(search(Formula), SFormula, SAdd) :-
	!,
	write(search),nl,
	set_srch_mode(direct),
	generate_cj_code(Formula, SFormula, SAdd),
	store_reified_variables_labeling(SAdd),
	unset_srch_mode.

% labeling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(labeling(Vars), STerm, SAdd) :-
	!,
	order_labelings([labeling(Vars)], OTerm),
	generate_cj_code(OTerm, LabelingTerm, SAdd),
	SolverIdent = s,
	STerm = (
		  LabelingTerm,
		  cj_obj_mthd_call(SolverIdent, generateSearchStrategy),
		  cj_obj_mthd_call(SolverIdent, launch),
		  cj_obj_static_mthd_call(cPSolver, flushLogs),
		  cj_obj_mthd_call('System.out', println(cj_obj_mthd_call(SolverIdent, pretty)))
		).

generate_cj_code(labeling([], Vars), STerm, SAdd) :-
	!,
	generate_cj_code(labeling([up], Vars), STerm, SAdd).
generate_cj_code(labeling([up], Vars), STerm, SAdd) :-
	!,
	generate_cj_code(Vars, SVars, SAdd),
	SolverIdent = s,
	(
	  first_labeling(true)
	->
	  retractall(first_labeling(_)),
	  assert(first_labeling(false)),
	  cj_solver_cstr(SCstrs),
	  list2seq(SCstrs, SolverConstraints),
	  (
	    cj_solver_goal(SolverGoals),
	    SolverGoals \== []
	  ->
	    Functor = addGoal,
	    list2seq(SolverGoals, AST)
	  ;	    	  	    
	    Functor = attachGoal,
	    AST = true
	  ),
	  SolverInit = (
			 cj_var_decl(cj_class(solver), SolverIdent) = cj_obj_new(cPSolver),
			 cj_obj_mthd_call(SolverIdent, read(cj_obj_field(cj_pkg_path([rules2cp,interpreter,java,r2cp]), model))),
			 cj_obj_static_mthd_call(cPSolver, setVerbosity(cj_obj_field('CPSolver', 'SEARCH'))),
			 SolverConstraints 
		       )
	;
	  Functor = addGoal,
	  AST = true,
	  SolverInit = true
	),
	ST =.. [Functor, cj_obj_new(assignVar(cj_obj_new(staticVarOrder(cj_obj_mthd_call(SolverIdent, getVar(cj_obj_new(cj_array_of(integerVariable), SVars))))),
					      cj_obj_new(increasingDomain)))],
	STerm = (SolverInit, AST, cj_obj_mthd_call(SolverIdent, ST)).
generate_cj_code(labeling([down], Vars), STerm, SAdd) :-
	!,	
	generate_cj_code(Vars, SVars, SAdd),	
	SolverIdent = s,
	(
	  first_labeling(true)
	->	  
	  retractall(first_labeling(_)),
	  assert(first_labeling(false)),
	  cj_solver_cstr(SCstrs),
	  list2seq(SCstrs, SolverConstraints),
	  (
	    cj_solver_goal(SolverGoals),
	    SolverGoals \== []
	  ->
	    Functor = addGoal,
	    list2seq(SolverGoals, AST)
	  ;	    	  	    
	    Functor = attachGoal,
	    AST = true
	  ),
	  SolverInit = (
			 cj_var_decl(cj_class(solver), SolverIdent) = cj_obj_new(cPSolver),
			 cj_obj_mthd_call(SolverIdent, read(cj_obj_field(cj_pkg_path([rules2cp,interpreter,java,r2cp]), model))),
			 cj_obj_static_mthd_call(cPSolver, setVerbosity(cj_obj_field('CPSolver', 'SEARCH'))),
			 SolverConstraints
		       )
	;
	  Functor = addGoal,
	  AST = true,
	  SolverInit = true
	),
	ST =.. [Functor, cj_obj_new(assignVar(cj_obj_new(staticVarOrder(cj_obj_mthd_call(SolverIdent, getVar(cj_obj_new(cj_array_of(integerVariable), SVars))))),
					      cj_obj_new(decreasingDomain)))],
	STerm = (SolverInit, AST, cj_obj_mthd_call(SolverIdent, ST)).
generate_cj_code(labeling([bisect], Vars), STerm, SAdd) :-
	!,
	generate_cj_code(Vars, SVars, SAdd),
	SolverIdent = s,
	(
	  first_labeling(true)
	->
	  retractall(first_labeling(_)),
	  assert(first_labeling(false)),	  
	  cj_solver_cstr(SCstrs),
	  list2seq(SCstrs, SolverConstraints),		  
	  (
	    cj_solver_goal(SolverGoals),
	    SolverGoals \== []
	  ->
	    Functor = addGoal,
	    list2seq(SolverGoals, AST)
	  ;	    	  	    
	    Functor = attachGoal,
	    AST = true
	  ),
	  SolverInit = (
			 cj_var_decl(cj_class(solver), SolverIdent) = cj_obj_new(cPSolver),
			 cj_obj_mthd_call(SolverIdent, read(cj_obj_field(cj_pkg_path([rules2cp,interpreter,java,r2cp]), model))),
			 cj_obj_static_mthd_call(cPSolver, setVerbosity(cj_obj_field('CPSolver', 'SEARCH'))),
			 SolverConstraints 
		       )
	;
	  Functor = addGoal,
	  AST = true,
	  SolverInit = true
	),
	ST =.. [Functor, cj_obj_new(dichotomicBranching(cj_obj_new(staticVarOrder(cj_obj_mthd_call(SolverIdent, getVar(cj_obj_new(cj_array_of(integerVariable), SVars)))))))],
	STerm = (SolverInit, AST, cj_obj_mthd_call(SolverIdent, ST)).

% optimization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(Term, STerm, SAdd) :-
	functor(Term, Functor, Arity),
	lang(optim_preds, OptimPreds),
	memberchk(Functor/Arity, OptimPreds),
	!,	
	(
	  Arity=:=2
	->
	  Term =.. [Functor, SearchTree, Cost]
	;
	  (
	    Arity=:=1
	  ->
	    Term =.. [Functor, Cost],
	    SearchTree = none
	  ;
	    raise_exception(code_generation(not_defined, [Term], []))
	  )
	),
	generate_cj_code(Cost, SCost, _),
	(
	  Functor == maximize
	->
	  CostLabelOpts = [down]
	;
	  (
	    Functor == minimize
	  ->
	    CostLabelOpts = [up]
	  ;
	    raise_exception(code_generation(not_defined, [Term], []))
	  )
	),		
	(
	  SearchTree == none
	->	  
	  BBGoal = true,
	  SAddST = []
	;
	  generate_cj_code(SearchTree, SSearchTree, SAddST),
	  BBGoal = SSearchTree
	),
	generate_cj_code(BBGoal, SBBGoal, SAddG),
	append(SAddST, SAddG, SAdd),
	SolverIdent = s,
	SolverInit = (
		       cj_var_decl(cj_class(solver), SolverIdent) = cj_obj_new(cPSolver),
		       cj_obj_mthd_call(SolverIdent, read(cj_obj_field(cj_pkg_path([rules2cp,interpreter,java,r2cp]), model))),
		       cj_obj_static_mthd_call(cPSolver, setVerbosity(cj_obj_field('CPSolver', 'SEARCH')))
		     ),	
	SOptTerm =.. [Functor, cj_obj_mthd_call(SolverIdent, getVar(SCost)), 'true'],

	cj_solver_goal(SolverGoals),
	list2conj(SolverGoals, AST),
	generate_cj_code(AST, AddSolverGoal, _),
	STerm = (
		  SBBGoal,
		  SolverInit,
		  AddSolverGoal,
		  cj_obj_mthd_call(SolverIdent, SOptTerm),
		  cj_obj_static_mthd_call(cPSolver, flushLogs),
		  cj_obj_mthd_call('System.out', println(cj_obj_mthd_call(SolverIdent, pretty)))
		).

% fdvar. domain constraint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(X in L,
		 domain(cj_obj_new(cj_array_of(integerVariable), SVars), LB, UB), SAdd) :-
	!,
	variables(X, Vars),
	generate_cj_list(Vars, SVars, SAdd1),
	generate_cj_range_list(L, LB..UB, SAdd2),
	append(SAdd1, SAdd2, SAdd).

% fdvar. domain constraint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(domain(VarsTerm, Min, Max),
		 domain(cj_obj_new(cj_array_of(integerVariable), SVars), VMin, VMax), SAdd) :- 
	!,
	variables(VarsTerm, Vars),
	generate_cj_list(Vars, SVars, SAdd1),
	generate_cj_list([Min, Max], [SMin, SMax], SAdd2),
	(
	  integer(SMin)
	->
	  VMin = SMin,
	  SAdd3 = SAdd1
	;	  
	  fresh_symbol('vmin', VMinVar),
	  VMin = cj_obj_mthd_call(VMinVar, getUppB),
	  append(SAdd1, [(cj_var_decl(cj_class(integerVariable), VMinVar) = makeIntVar(cj_string(VMinVar), 0, 10000)),
			 eq(VMinVar, SMin)], SAdd3)
	),
	(
	  integer(SMax)
	->
	  VMax = SMax,
	  SAdd4 = SAdd3
	;	  
	  fresh_symbol('vmax', VMaxVar),
	  VMax = cj_obj_mthd_call(VMaxVar, getUppB),
	  append(SAdd3, [(cj_var_decl(cj_class(integerVariable), VMaxVar) = makeIntVar(cj_string(VMaxVar), 0, 10000)),
			 eq(VMaxVar, SMax)], SAdd4)
	),
	append(SAdd2, SAdd4, SAdd).

% fdvar. domain lower bound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(domain_min(Var), cj_obj_mthd_call(SVar, getLowB), SAdd) :-
	!,
	generate_cj_code(Var, SVar, SAdd).

% fdvar. domain upper bound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(domain_max(Var), cj_obj_mthd_call(SVar, getUppB), SAdd) :-
	!,
	generate_cj_code(Var, SVar, SAdd).

% fdvar. domain cardinality
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(domain_size(Var), cj_obj_mthd_call(SVar, getDomainSize), SAdd) :-
	!,
	generate_cj_code(Var, SVar, SAdd).

% global constraint all_different
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(all_different(Vars),
		 allDifferent(cj_obj_new(cj_array_of(integerVariable), SVars)),
		 SAdd) :-
	!,	
	generate_cj_code(Vars, SVars, SAdd).

% % global constraint disjoint2
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate_cj_code(disjoint2(Xs, Ys, Ss), disjoint2(Data), []) :-
% 	!,
% 	generate_cj_code([Xs, Ys, Ss], SArgs, _),
% 	append(SArgs, [Data], DisjDataArgs),
% 	DisjDataTerm =.. [disjoint_data | DisjDataArgs],
% 	call(DisjDataTerm).

% global constraint cumulative
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(cumulative(Os, Ds, Hs, L),
		 cumulative(cj_obj_new(cj_array_of(integerVariable), SOs),
			    cj_obj_new(cj_array_of(integerVariable), SDs),
			    cj_obj_new(cj_array_of(integerVariable), SHs), SL),
		SAdd) :-
	!,
	generate_cj_list([Os, Ds, Hs, L], [SOs, SDs, SHs, SL], SAdd).

% global constraint geost
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(geost_constraint(PreObjects, PreSBoxes, Dims),	STerm, []) :-
	!,
	generate_cj_geost_object(PreObjects, Dims, Objects, _Oids, ObjectsShapeDoms),	
	store_term(ObjectsShapeDoms, cstr_terms),
	generate_cj_geost_sbox(PreSBoxes, Sboxs),
	remove_dups(Sboxs, SBoxes0),		
	list2seq(SBoxes0, SBoxes),	
	Decl = (
		 cj_var_decl(cj_generic(vector, geostObject), geost_obj) = cj_obj_new(cj_generic(vector, geostObject)),
		 cj_var_decl(cj_generic(vector, shiftedBox), geost_sb) = cj_obj_new(cj_generic(vector, shiftedBox))
	       ),
	STerm = (
		  Decl,
		  Objects,
		  SBoxes,
		  geost(Dims, geost_obj, geost_sb, non_overlapping(Dims, geost_obj))
		).

% atomic 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(Term, STerm, []) :-
	atomic(Term),
	dict_cj(Term, STerm),
	!.
generate_cj_code(Term, Term, []) :-
	atomic(Term),
	!.

% list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(Term, STerm, SAdd) :-
	is_list(Term),
	!,
	generate_cj_list(Term, STerm, SAdd).

% arithmetic 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(Term, STerm, SAdd) :-	
	is_arithop(Term),
	Term =.. [Op | Args],
	contains_compop(Args),
	!,
	reify_cj_compop(Args, RArgs, SAdd),
	dict_cj(Op, TOp),
	ST =.. [TOp | RArgs],
	eval(ST, STerm).

% search (label) node
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(node(_L, Term), STerm, SAdd) :-
	!,
	generate_cj_code(Term, STerm, SAdd).

% user output 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(write(Term), STerm, SAdd) :-	
	!,
	generate_cj_code(Term, STerm, SAdd),
	format("\"--- User Output ---\"\n ~w ", [Term]),
	format("\n \"~~>\"\n",[]),
	format(" ~w\n\"--- End User Output ---\"\n", [STerm]).

% arithmetic negation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(Term, STerm, SAdd) :-
	Term =.. [-, Arg],
	generate_cj_list(Arg, SArg, SAdd),
	ST =.. [neg, SArg],
	eval(ST, STerm).

% default 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_code(P or Q, STerm, [cj_var_decl(cj_class(integerVariable), ReifVar) = makeIntVar(cj_string(ReifVar), 0, 1) | SAdd]) :-
	srch_mode(_),	
	!,
	fresh_symbol('reif', ReifVar),
	generate_cj_code((ReifVar = 1 equiv P), SP, AddP),
	generate_cj_code((ReifVar = 0 equiv Q), SQ, AddQ),
	append(AddP, AddQ, SAdd),
	STerm = (SP and SQ).

generate_cj_code(Term, STerm, SAdd) :-
	Term =.. [Op | Args],
	dict_cj(Op, TOp),
	!,
	generate_cj_list(Args, SArgs, SAdd),
	ST =.. [TOp | SArgs],
	eval(ST, STerm).

generate_cj_code(Term, STerm, SAdd) :-
	Term =.. [Op | Args],
	generate_cj_list(Args, SArgs, SAdd),
	ST =.. [Op | SArgs],
	eval(ST, STerm).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% data-structure-dedicated code generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_list([], [], []).
generate_cj_list([UT | UTerms], [ST | STerms], SAdd) :-
	generate_cj_code(UT, ST, SA),
	generate_cj_list(UTerms, STerms, SAs),
	append(SA, SAs, SAdd).

% record 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_record([], [], []).
generate_cj_record([Key = Val | T], [Key = SVal | ST], SAdd) :-
	generate_cj_code(Val, SVal, SA),
	generate_cj_record(T, ST, SAs),
	append(SA, SAs, SAdd).

% range list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_cj_range_list([L .. U], (L .. U), []) :-
	integer(L),
	integer(U),
	!.
generate_cj_range_list([L .. U], (L .. U), []) :-
	L == inf,
	integer(U),
	!.
generate_cj_range_list([L .. U], (L .. U), []) :-
	integer(L),
	U == sup,
	!.
generate_cj_range_list([S], {S}, []) :- 
	!,
	integer(S).
generate_cj_range_list([L .. U | T], RUnion \/ (L..U), SAdd) :-
	!,
	integer(L),
	integer(U),
	generate_cj_range_list(T, RUnion, SAdd).
generate_cj_range_list([S | T], RUnion \/ {S}, SAdd) :-	
	integer(S),
	generate_cj_range_list(T, RUnion, SAdd).


% explicit reification 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
reify_cj_compop(Term, RTerm, Add) :- 
	is_compop(Term),
	!,
	Term =.. [Op | Args],
	reify_cj_compop(Args, RArgs, As),
	dict_cj(Op, TOp),
	ReifTerm =.. [TOp | RArgs],	
	fresh_symbol('reif', ReifVar),
	RTerm = ReifVar,
	AddTerm =.. [reifiedIntConstraint, ReifVar, ReifTerm],
	append(As, [cj_var_decl(cj_class(integerVariable), ReifVar) = makeIntVar(cj_string(ReifVar), 0, 1),
		    AddTerm],
	       Add).
reify_cj_compop([], [], []) :-
	!.
reify_cj_compop([H | T], [RT | RTs], Add) :-	
	reify_cj_compop(H, RT, A),
	reify_cj_compop(T, RTs, As),
	!,
	append(A, As, Add).
reify_cj_compop(Term, RTerm, Add) :-
	generate_cj_code(Term, RTerm, Add).


% place variable declaration first
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
cj_var_decl_first(InstrList, VarDeclFirstInstrList) :-
	cj_var_decl_first(InstrList, VarDeclList, OtherInstrList),
	append(VarDeclList, OtherInstrList, VarDeclFirstInstrList).

cj_var_decl_first([], [], []).
cj_var_decl_first([cj_var_decl(T, N) = V | InstrList],
		  [cj_var_decl(T, N) = V | VarDeclList], OtherInstrList) :-
	!,
	cj_var_decl_first(InstrList, VarDeclList, OtherInstrList).
cj_var_decl_first([Instr | InstrList], VarDeclList, [ Instr | OtherInstrList]) :-
	cj_var_decl_first(InstrList, VarDeclList, OtherInstrList).


% geost  
%%%%%%%%%%%%%%%%%%%%
generate_cj_geost_object([], _, true, [], []).
generate_cj_geost_object([PreObj | PreObjs], Dims, (Obj, Objs), [Oid | Oids], ObjShpDoms) :-
	generate_cj_geost_object(PreObjs, Dims, Objs, Oids, OSDoms),

	PreObj = [Oid, S, Sids, Origins],
	atomize([start, Oid], '_', OStart),
	atomize([duration, Oid], '_', ODuration),
	atomize([end, Oid], '_', OEnd),
	atomize([sid, Oid], '_', OSid),
	atomize([values, Oid], '_', OValues),	
	(
	  is_var(S)
	->
	  SID = (
		  (cj_var_decl(cj_array_of(int), OValues) = cj_obj_new(cj_array_of(int), Sids)),
		  (cj_var_decl(cj_class(integerVariable), OSid) = makeIntVar(cj_string(OValues), OValues, cj_string('cp:enum')))		 
		),
	  
	  generate_cj_code(S, ShapeVar, _),
	  Cstr = cj_obj_mthd_call(s, post(cj_obj_new(element(cj_obj_mthd_call(s, getVar(ShapeVar)),
							     OValues,
							     cj_obj_mthd_call(s, getVar(OSid)),
							     -1)))),
	  cj_solver_cstr(SCstrs),
	  append(SCstrs, [Cstr], SolverCstrs),
	  retractall(cj_solver_cstr(_)),
	  assert(cj_solver_cstr(SolverCstrs))
	;
	  nth1(S, Sids, Sid),
	  ObjShpDoms = OSDoms,
	  SID = (
		  cj_var_decl(cj_class(integerVariable), OSid) = makeIntVar(cj_string(OSid), Sid, Sid)
		)
	),
	generate_cj_code(Origins, OOrigins, _),		
	Start = (cj_var_decl(cj_class(integerVariable), OStart) = makeIntVar(cj_string(OStart), 0, 0)),
	Dur = (cj_var_decl(cj_class(integerVariable), ODuration) = makeIntVar(cj_string(ODuration), 1, 1)),
	End = (cj_var_decl(cj_class(integerVariable), OEnd) = makeIntVar(cj_string(OEnd), 1, 1)),		
	Obj = (
		Start,
		Dur,
		End,
		SID,
		cj_obj_mthd_call(geost_obj, add(cj_obj_new(geostObject(Dims, Oid, OSid,
								       cj_obj_new(cj_array_of(integerVariable), OOrigins),
								       OStart, ODuration, OEnd))))
	      ).

generate_cj_geost_sbox([], []).
generate_cj_geost_sbox([PreSbox | PreSboxes], Sboxes) :-
	generate_cj_geost_sbox(PreSboxes, Sboxs),
	traverse_cj_presbox_shapes(PreSbox, Sbox),
	append(Sbox, Sboxs, Sboxes).

traverse_cj_presbox_shapes([],[]).
traverse_cj_presbox_shapes([PreSboxShape | PreSboxShapes], SboxesShapes) :-
	traverse_cj_presbox_shapes(PreSboxShapes,SboxShapes),
	traverse_cj_presbox_boxes(PreSboxShape, Sboxes),
	append(Sboxes, SboxShapes, SboxesShapes).

traverse_cj_presbox_boxes([],[]).
traverse_cj_presbox_boxes([PreSboxBox | PreSboxBoxes],[Sbox | Sboxes]) :-
	traverse_cj_presbox_boxes(PreSboxBoxes,Sboxes),
	PreSboxBox = [Sid, Offset, Sizes],
	Sbox = cj_obj_mthd_call(geost_sb, add(cj_obj_new(shiftedBox(Sid,
								    cj_obj_new(cj_array_of(int), Offset),
								    cj_obj_new(cj_array_of(int), Sizes))))).


store_reified_variables_labeling(AddCode) :-
	filter_reified_variables_labeling(AddCode, ReifVars),
	SolverIdent = s,
	SGoal = cj_obj_mthd_call(SolverIdent,
				 attachGoal(cj_obj_new(assignVar(
								 cj_obj_new(staticVarOrder(cj_obj_mthd_call(SolverIdent, getVar(cj_obj_new(cj_array_of(integerVariable), ReifVars))))),
								 cj_obj_new(decreasingDomain)
								)
						      )
					   )
				),
	
	cj_solver_goal(SGoals),
	append(SGoals, [SGoal], SolverGoals),
	retractall(cj_solver_goal(_)),
	assert(cj_solver_goal(SolverGoals)),
	
	write_list(SolverGoals),nl.
	
filter_reified_variables_labeling([], []) :-
	!.
filter_reified_variables_labeling([(cj_var_decl(_Type, Name) = makeIntVar(cj_string(Name),0,1)) | T],
				  [Name | R]) :-
	!,
	filter_reified_variables_labeling(T, R).
filter_reified_variables_labeling([_ | T], R) :-
	filter_reified_variables_labeling(T, R).

