% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file ocg.sp.pl
%

% CLPFD SICStus Prolog operators dictionary 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
dict_sp(<, #<).
dict_sp(=<, #=<).
dict_sp(=, #=).
dict_sp(#, #\=).
dict_sp(>=, #>=).
dict_sp(>, #>).
dict_sp(true,  1).
dict_sp(false, 0).
dict_sp(and, #/\).
dict_sp(or, #\/).
dict_sp(not, #\).
dict_sp(equiv, #<=>).
dict_sp(implies, #=>).
dict_sp(T, T).

dict_sp_dyn(<, fd_lt).
dict_sp_dyn(=<, fd_leq).
dict_sp_dyn(=, fd_eq).
dict_sp_dyn(#, fd_neq).
dict_sp_dyn(>=, fd_geq).
dict_sp_dyn(>, fd_gt).
dict_sp_dyn(true,  1).
dict_sp_dyn(false, 0).
dict_sp_dyn(and, fd_and).
dict_sp_dyn(or, fd_or).
dict_sp_dyn(not, fd_not).
dict_sp_dyn(equiv, fd_equiv).
dict_sp_dyn(implies, fd_implies).
dict_sp_dyn(T, T).

% prolog {cons, conjunction, disjunction}  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code('.', '\'.\'', []) :-
	!.

generate_sp_code(',', '\',\'', []) :-
	!.

generate_sp_code(';', '\';\'', []) :-
	!.

% integer
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, Term, []) :-	
	integer(Term),
	!.

% -2147483616
generate_sp_code(min_integer, -268435456, []) :-
	!.
% 2147483616
generate_sp_code(max_integer, 267999999, []) :-
	!.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% dynamic code generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% free variable specialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(free_var(X, N, I), Var, [rcp_var(from(Y, N, I), Var)]) :-
	cg_mode(dynamic),
	!,
	delete_uids(X, Y),
	serialize_term(Y, SX),
	to_upper_case(SX, SXUpp),
	serialize_int_var_infx(N, NI),
	serialize_int_var_sufx(I, SI),
	atomize([SXUpp, NI, SI], '_', Var).

% variable specialization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(var(X, I), STerm, []) :-
	cg_mode(dynamic),
	!,		
	serialize_term(X, SX),
	to_upper_case(SX, SXUpp),
	serialize_int_var_sufx(I, SI),
	atomize([SXUpp, SI], '_', STerm).
generate_sp_code(var(X, I), _STerm, []) :- 
	!,
	raise_exception(code_generation(not_bounded_variable, [var(X, I)], [])).

generate_sp_code(Term, Term, []) :-
	cg_mode(dynamic),
	is_var_name(Term),
	!.

% % dynamic/search atomic (function)
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate_sp_code(Term, V, [STerm]) :-
% 	cg_mode(dynamic),
% 	srch_mode(_),
% 	atomic(Term),
% 	constant_function(Term, _Head, _Body),
% 	!,
% 	atomize([rcp, Term, search], '_', Functor),
% 	fresh_local_var_name(V),	
% 	STerm =.. [Functor, V].

% % dynamic atomic (function)
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% generate_sp_code(Term, V, [STerm]) :-
% 	cg_mode(dynamic),
% 	atomic(Term),
% 	constant_function(Term, _Head, _Body),
% 	!,
% 	atomize([rcp, Term], '_', Functor),
% 	fresh_local_var_name(V),
% 	STerm =.. [Functor, V].

% dynamic/search atomic (relation)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, [STerm]) :-
	cg_mode(dynamic),
	srch_mode(_),
	atomic(Term),
	constant_relation(Term, _Head, _Body),
	!,
	atomize([rcp, Term, search], '_', Functor),	
	fresh_local_var_name(V),
	STerm =.. [Functor, V].

% dynamic atomic (relation)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, [STerm]) :-
	cg_mode(dynamic),
	atomic(Term),
	constant_relation(Term, _Head, _Body),
	!,
	atomize([Term], '_', Functor),
	fresh_local_var_name(V),
	STerm =.. [Functor, V].

% dynamic atomic built-in operator
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, []) :-
	cg_mode(dynamic),
	atomic(Term),
	dict_sp_dyn(Term, STerm),
	!.

% dynamic atomic default
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, Term, []) :-
	cg_mode(dynamic),
	atomic(Term),
	!.

% dynamic length/2 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(length(Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Term, STerm, SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [length(STerm, V)], SAdd).

% dynamic nth/2 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(nth(N, Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list([N, Term], [SN, STerm], SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [nth1(SN, STerm, V)], SAdd).

% dynamic pos/2 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(pos(Elem, Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list([Elem, Term], [SElem, STerm], SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [nth1(V, STerm, SElem)], SAdd).

% dynamic/search comparison (<|=<|=|>=|>|#)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_),
	is_compop(Term),
	!,
	Term =.. [Op | Args],
	dict_sp_dyn(Op, TOp),
	generate_sp_list(Args, SArgs, SAdd),
	STerm =.. [TOp | SArgs].

% dynamic comparison (<|=<|=|>=|>|#)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),
	is_compop(Term),
	!,
	Term =.. [Op | Args],
	dict_sp_dyn(Op, TOp),
	generate_sp_list(Args, SArgs1, SAdd1),
	fresh_local_var_name(V),
	append(SArgs1, [V], SArgs),
	STerm =.. [TOp | SArgs],				
	append(SAdd1, [STerm], SAdd).

% dynamic arithmetic exp/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(exp(X, E), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list([X, E], [SX, SE], SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [exp(SX, SE, V)], SAdd).

% dynamic arithmetic log/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(log(B, X), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list([B, X], [SB, SX], SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [log(SB, SX, V)], SAdd).

% dynamic arithmetic (+|-|*|/|min|max|abs)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),
	is_arithop(Term),
	!,
	Term =.. [Functor | Args],
	generate_sp_list(Args, SArgs, SAdd1),
	fresh_local_var_name(V),
	dict_sp_dyn(Functor, SFunctor),
	STerm =.. [SFunctor | SArgs],
	append(SAdd1, [fd_aeq(STerm, V)], SAdd).

% dynamic/search not elimination
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(not (P and Q), S, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!,
	generate_sp_code((not P or not Q), S, SAdd).
generate_sp_code(not (P or Q), S, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!,
	generate_sp_code((not P and not Q), S, SAdd).
generate_sp_code(not not P, S, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!,
	generate_sp_code(P, S, SAdd).
generate_sp_code(not P, S, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	is_compop(P),
	!,
	P =.. [F | Args],
	compop_neg(F, NF),
	NP =.. [NF | Args],
	generate_sp_code(NP, S, SAdd).

% dynamic/search disjunction (or)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(P or Q, STerm, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!,	
	generate_sp_code(P, SP, SAddP),
	generate_sp_code(Q, SQ, SAddQ),	
	(
	  is_var_name(SP)
	->
	  append(SAddP, [rcp_assert(SP)], SAddP0),
	  list2seq(SAddP0, SPTerm),
	  SAdd0 = []	
	;
	  SPTerm = SP,
	  SAdd0 = SAddP
	),
	(
	  is_var_name(SQ)
	->
	  append(SAddQ, [rcp_assert(SQ)], SAddQ0),
	  list2seq(SAddQ0, SQTerm),
	  SAdd = SAdd0	
	;
	  SQTerm = SQ,
	  append(SAdd0, SAddQ, SAdd)
	),
	STerm = ';'(SPTerm, SQTerm).

% dynamic/search conjunction (and)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(P and Q, STerm, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),	
	!,
	generate_sp_code(P, SP, SAddP),
	generate_sp_code(Q, SQ, SAddQ),	
				% write(srch(rcp_and(SP, SQ))),nl,	
	(
	  is_var_name(SP)
	->
	  append(SAddP, [rcp_assert(SP)], SAddP0),
	  list2seq(SAddP0, SPTerm),
	  SAdd0 = []
	;
	  SPTerm = SP,
	  SAdd0 = SAddP
	),
	(
	  is_var_name(SQ)
	->
	  append(SAddQ, [rcp_assert(SQ)], SAddQ0),
	  list2seq(SAddQ0, SQTerm),
	  SAdd = SAdd0
	;
	  SQTerm = SQ,
	  append(SAdd0, SAddQ, SAdd)
	),
	STerm = ','(SPTerm, SQTerm).

% dynamic/search implication (implies)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(P implies Q, S, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_SMode),	
	!,	
	generate_sp_code(P, SP, SAddP),
	generate_sp_code(not P, SNP, SAddNP),
	generate_sp_code(Q, SQ, SAddQ),
				%write(srch(rcp_implies(SP, SQ))),nl,
	list2seq(SAddQ, SAddQSeq),
	fresh_local_var_name(S),
	STerm = (
		  SP,
		  SAddQSeq,
		  fd_equiv(SQ, S, 1)
		;
		  SNP,
		  fd_aeq(S, 1)
		),
	append(SAddP, SAddNP, SAddPNP),
	append(SAddPNP, [STerm], SAdd).

% dynamic conjunction or disjunction (and|or)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(P and Q, ','(SP, SQ), SAdd) :-
	cg_mode(dynamic),
	is_inlined(P),
	is_inlined(Q),
	!,
	generate_sp_code(P, SP, SAddP),
	generate_sp_code(Q, SQ, SAddQ),
	append(SAddP, SAddQ, SAdd).
generate_sp_code(P and Q, SQ, SAdd) :-
	cg_mode(dynamic),
	is_inlined(P),
	!,
	generate_sp_code(P, SP, SAddP1),
	generate_sp_code(Q, SQ, SAddQ),
	append(SAddP1, [SP], SAddP),
	append(SAddP, SAddQ, SAdd).
generate_sp_code(P and Q, SP, SAdd) :-
	cg_mode(dynamic),
	is_inlined(Q),
	!,
	generate_sp_code(P, SP, SAddP),
	generate_sp_code(Q, SQ, SAddQ1),
	append(SAddQ1, [SQ], SAddQ),
	append(SAddP, SAddQ, SAdd).

% dynamic/search logic connectives (not|equiv)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_),
	is_logop(Term),
	!,
	Term =.. [Functor | Args],
	dict_sp_dyn(Functor, SFunctor),
	generate_sp_list(Args, SArgs1, SAdd1),
	fresh_local_var_name(V),
	append(SArgs1, [V], SArgs),
	STerm =.. [SFunctor | SArgs],
	append(SAdd1, [STerm], SAdd).

% dynamic logic connectives (and|or|not|equiv|implies)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),	
	is_logop(Term),	
	!,
	Term =.. [Functor | Args],
	dict_sp_dyn(Functor, SFunctor),
	generate_sp_list(Args, SArgs1, SAdd1),
	fresh_local_var_name(V),
	append(SArgs1, [V], SArgs),
	STerm =.. [SFunctor | SArgs],
		
				%write(dyn(logop(STerm))),nl,
	
	append([STerm], SAdd1, SAdd).

% dynamic variables/1
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(variables(Term), V, SAdd) :-
	cg_mode(dynamic),	
	!,
	generate_sp_code(Term, STerm, SAdd1),
	fresh_local_var_name(V),
	append(SAdd1, [variables(STerm, V)], SAdd).

% dynamic let/3 (binder)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(let(U, V, E), STerm, SAdd) :-
	cg_mode(dynamic),	
	!,
	generate_sp_list([U, V], [SU, SV], SAdd0),
	subst(SV, SU, SAdd0, SAdd1),
				% append(SAdd1, [SU = SV], SAdd0),
	generate_sp_code(E, STerm, SAdd2),
	append(SAdd1, SAdd2, SAdd).

% dynamic write/1
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(rcp_write(Term), rcp_write(STerm), SAdd) :-
	cg_mode(dynamic),	
	!,
	generate_sp_code(Term, STerm, SAdd).

% labeling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(labeling(Term), rcp_labeling(STerm), SAdd) :-
	cg_mode(dynamic),	
	!,
	generate_sp_code(Term, STerm, SAdd).

% dynamic optim. (minimize/1)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(minimize(CostTerm),
		 minimize(rcp_labeling([SCostTerm]), SCostTerm),
		 SAdd) :-
 	cg_mode(dynamic),
 	!,	
 	generate_sp_code(CostTerm, SCostTerm, SAdd).

% dynamic optim. (minimize/2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(minimize(SearchTerm, CostTerm),
		 minimize((SSearchTerm, rcp_labeling([SCostTerm])), SCostTerm),
		 SAdd) :-
 	cg_mode(dynamic),
 	!,	
 	generate_sp_code(SearchTerm, SSearchTerm1, SAddSearch1),	
	(	  
	  is_var_name(SSearchTerm1)
	->
	  append([rcp_assert(SSearchTerm1)], SAddSearch1, SAdd0),
	  list2seq(SAdd0, SSearchTerm)
	;	  
	  SAdd0 = SAddSearch1,
	  list2seq(SAdd0, STerm0),
	  SSearchTerm = (STerm0, SSearchTerm1)
	),
 	generate_sp_code(CostTerm, SCostTerm, SAdd).

% dynamic optim. (maximize/1)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(maximize(CostTerm),
		 maximize(rcp_labeling([SCostTerm]), SCostTerm),
		 SAdd) :-
 	cg_mode(dynamic),
 	!,	
 	generate_sp_code(CostTerm, SCostTerm, SAdd).

% dynamic optim. (maximize/2)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(maximize(SearchTerm, CostTerm),
		 maximize((SSearchTerm, rcp_labeling([SCostTerm])) ,SCostTerm),
		 SAdd) :-
 	cg_mode(dynamic),
 	!,	
 	generate_sp_code(SearchTerm, SSearchTerm1, SAddSearch1),	
	(	  
	  is_var_name(SSearchTerm1)
	->
	  append([rcp_assert(SSearchTerm1)], SAddSearch1, SAdd0),
	  list2seq(SAdd0, SSearchTerm)
	;	  
	  SAdd0 = SAddSearch1,
	  list2seq(SAdd0, STerm0),
	  SSearchTerm = (STerm0, SSearchTerm1)
	),
 	generate_sp_code(CostTerm, SCostTerm, SAdd).

% dynamic attribute
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(rcp_att(Functor, [Arg]), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Arg, SArg, SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [rcp_att(Functor, SArg, V)], SAdd).

% dynamic range
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code([L .. U], V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list([L, U], [SL, SU], SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [rcp_range([SL..SU], V)], SAdd).

% dynamic "inplace" cons
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(cons(HTerm, TTerm), [SHTerm | STTerm], SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(HTerm, SHTerm, SAdd1),
	generate_sp_code(TTerm, STTerm, SAdd2),
	append(SAdd1, SAdd2, SAdd).

% dynamic "inplace" list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(list(Term), STerm, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_list(Term, STerm, SAdd).

% dynamic list
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	is_list(Term),
	!,
	generate_sp_list(Term, STerm, SAdd).

% dynamic record
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(rcp_rec(Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_record(Term, STerm, SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [rcp_rec(STerm, V)], SAdd).

% dynamic search 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(search(Term), STerm, SAdd) :-
	cg_mode(dynamic),
	!,
        write(dynamic(search)),nl,
	set_srch_mode(reordering),	
	generate_sp_code(Term, STerm, SAdd),
	unset_srch_mode.

% dynamic {variable,value}_ordering/1
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),	
	functor(Term, Functor, Arity),
	lang(vordercrit_preds, OrderPreds),
	memberchk(Functor/Arity, OrderPreds),
	!,
	Term =.. [Functor | Args],
	generate_sp_code(list(Args), SArgs, SAdd),
	STerm =.. [Functor | SArgs].

% dynamic {greatest,least}/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	functor(Term, Functor, Arity),
	lang(order_preds, OrderPreds),
	memberchk(Functor/Arity, OrderPreds),
	!,
	Term =.. [Functor | [Args]],
	generate_sp_code(list(Args), SArgs, SAdd),
	STerm =.. [Functor | [SArgs]].

% dynamic {var,val}_order_criterion/2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	Term =.. [Functor | Args],
	length(Args, Arity),
	lang(dynvordercrit_preds, Preds),
	memberchk(Functor/Arity, Preds),
	!,
	generate_sp_list(Args, SArgs, SAdd),
	atomize([rcp, Functor], '_', SFunctor),
	STerm =.. [SFunctor | SArgs].

% dynamic labeling 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	Term =.. [Functor | Args],
	length(Args, Arity),
	lang(label_preds, LabelPreds),
	memberchk(Functor/Arity, LabelPreds),
	!,
	generate_sp_list(Args, SArgs, SAdd),
	STerm =.. [Functor | SArgs].

% dynamic global constraint all_different
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(all_different(Variables), all_different(SVariables), SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Variables, SVariables, SAdd).

% dynamic global constraint lexicographic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(lexicographic(ListOfListVariables),
		 lex_chain(SListOfListVariables), SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(ListOfListVariables, SListOfListVariables, SAdd).

generate_sp_code(lexicographic_strict(ListOfListOfVars),
		 lex_chain(SListOfListOfVars, [op(#<)]), SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(ListOfListOfVars, SListOfListOfVars, SAdd).

% dynamic global constraint cumulative
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(cumulative(Os, Ds, Hs, L),
		 cumulative(Tasks, [limit(GL), global(true)]), SAdd) :-
	cg_mode(dynamic),        
	!,
	generate_sp_code([Os, Ds, Hs, L], [GOs, GDs, GHs, GL], SAdd1),
	fresh_local_var_name(V),
	fresh_local_var_name(Vs),
	fresh_local_var_name(Tasks),	
	append(SAdd1,
	       [length(GOs, V),
		rcp_gen_vars(V, Vs),
		rcp_cumulative_tasks(GOs, GDs, GHs, Vs, Tasks),
		domain(Vs, 0, 1000)],
	       SAdd).

% dynamic global constraint disjoint2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(disjoint2(Xs, Ys, Ss), disjoint2(Data), SAdd) :-
	cg_mode(dynamic),
	!,	
	generate_sp_code([Xs, Ys, Ss], [GXs, GYs, GSs], SAdd1),
	fresh_local_var_name(Data),
	append(SAdd1, [rcp_disjoint2_data(GXs, GYs, GSs, Data)], SAdd).

% dynamic domain
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(V in D, SV in SD, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(V, SV, SAddV),
	generate_sp_code(D, SD, SAddD),
	append(SAddV, SAddD, SAdd).

generate_sp_code(Term, STerm, SAdd) :-
	cg_mode(dynamic),
	Term =.. [Functor | Args],
	is_domain(Term),
	!,
	generate_sp_list(Args, SArgs, SAdd),
	STerm =.. [Functor | SArgs].

% dynamic domain lower bound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_min(Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Term, STerm, SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [fd_min(STerm, V)], SAdd).

% dynamic domain upper bound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_max(Term), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Term, STerm, SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [fd_max(STerm, V)], SAdd).

% dynamic domain cardinality
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_size(Term), V, SAdd) :-
	cg_mode(dynamic),	
	!,
	generate_sp_code(Term, STerm, SAdd1),
	fresh_local_var_name(V),
 	append(SAdd1, [fd_size(STerm, V)], SAdd).

% dynamic dynamic (idempotence)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(dynamic(Term), STerm, SAdd) :-
	cg_mode(dynamic),
	!,	
	generate_sp_code(Term, STerm, SAdd).

% dynamic/search default 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),
	srch_mode(_),
	!,
				%write(Term),nl,	
	Term =.. [Op | Args],
	generate_sp_list(Args, SArgs, SAdd1),
	dict_sp_dyn(Op, TOp),
	(
	  (
	    is_builtinop(Op)
	  ;
	    is_inlined(Term)
	  )
	->
	  atomize([rcp, TOp], '_', Functor)
	;	  
	  atomize([rcp, TOp, search], '_', Functor)
	),
	
	
	fresh_local_var_name(V),
	append(SArgs, [V], VSArgs),
	STerm =.. [Functor | VSArgs],
	append(SAdd1, [STerm], SAdd).

% dynamic Prolog sequence and choice
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(';'(P, Q), ';'(P, Q), []) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!.
generate_sp_code(','(P, Q), ','(P, Q), []) :-
	cg_mode(dynamic),
	srch_mode(_SMode),
	!.

generate_sp_code(','(P, Q), ','(P, Q), []) :-
	cg_mode(dynamic),
	!.
generate_sp_code(';'(P, Q), ';'(P, Q), []) :-
	cg_mode(dynamic),
	!.
generate_sp_code('.'(P, Q), '.'(P, Q), []) :-
	cg_mode(dynamic),
	!.

% dynamic term call
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(rcp_call(Arg), V, SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(Arg, SArg, SAddArg),
	fresh_local_var_name(V),
	append([SArg], [V], VSArgs),
	STerm =.. [rcp_call | VSArgs],
	append(SAddArg, [STerm], SAdd).

% dynamic term default
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, V, SAdd) :-
	cg_mode(dynamic),
	!,
	Term =.. [Functor | Args],
	generate_sp_list(Args, SArgs, SAdd1),
	fresh_local_var_name(V),
	append(SArgs, [V], VSArgs),
	atomize([rcp, Functor], '_', TFunctor),
	STerm =.. [TFunctor | VSArgs],		
				% write(dyn(default(STerm))),nl,				
	append(SAdd1, [STerm], SAdd).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% static code generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% atom 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, []) :-	
	atom(Term),
	atom_codes(Term, [0'$ | T]),
	!,
	atom_codes(STerm, T).

% variable 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(free_var(X, N, I), Var, [rcp_var(from(Y, N, I), Var)]) :- 	
	!,
	delete_uids(X, Y),
	serialize_term(Y, SX),
	to_upper_case(SX, SXUpp),
	serialize_int_var_infx(N, NI),
	serialize_int_var_sufx(I, SI),
	atomize([SXUpp, NI, SI], '_', Var).

% dynamic tag
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(dynamic(Term), STerm, []) :-
	!,
	cg_mode(CGModeBefore),
	set_cg_mode(dynamic),
        	write(dynamic),nl,
	generate_sp_code(Term, STerm1, SAdd1),
			
	(	  
	  is_var_name(STerm1)
	->
	  append(SAdd1, [rcp_assert(STerm1)], SAdd),
	  list2seq(SAdd, STerm)
	;	  
	  list2seq(SAdd1, STerm0),
	  STerm = (STerm0, STerm1)
	),
	set_cg_mode(CGModeBefore).

% search
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(search(Formula), SFormula, SAdd) :-
	!,
	set_srch_mode(direct),
	subst_and_or(Formula, FormulaSubst),
	generate_sp_code(FormulaSubst, SFormula, SAdd),
	unset_srch_mode.

% labeling
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(labeling(Vars), STerm, SAdd) :-
	!,
        write(labeling),nl,
	order_labelings([labeling(Vars)], OTerm),
	generate_sp_code(OTerm, STerm, SAdd).

% optimization
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	functor(Term, Functor, Arity),
	lang(optim_preds, OptimPreds),
	memberchk(Functor/Arity, OptimPreds),
	!,
	(
	  Arity=:=2
	->
	  Term =.. [Functor, SearchTree, Cost]
	;
	  (
	    Arity=:=1
	  ->
	    Term =.. [Functor, Cost],
	    SearchTree = none
	  ;
	    raise_exception(code_generation(not_defined, [Term], []))
	  )
	),
	variables_set(Cost, CostVariables),
	generate_sp_code(CostVariables, SCostVariables, _),
	generate_sp_code(Cost, SCost, _),
	(
	  Functor == maximize
	->
	  CostLabelOpts = [down]
	;
	  (
	    Functor == minimize
	  ->
	    CostLabelOpts = [up]
	  ;
	    raise_exception(code_generation(not_defined, [Term], []))
	  )
	),
	CostLabeling = labeling(CostLabelOpts, SCostVariables),	
	(
	  SearchTree == none
	->
	  BBGoal = (CostLabeling)
	;
	  BBGoal = (SearchTree, CostLabeling)
	),
	generate_sp_code(BBGoal, SBBGoal, SAdd),
	STerm =.. [Functor, SBBGoal, SCost].

% fdvar. domain constraint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(V in D, SV in SD, SAdd) :-	
	!,
	generate_sp_code(V, SV, SAddV),
	generate_sp_range_list(D, SD, SAddD),
	append(SAddV, SAddD, SAdd).

% fdvar. domain constraint
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain(VarsTerm, Min, Max), domain(SVars, VMin, VMax), SAdd) :- 	
	!,
	variables(VarsTerm, Vars),
	generate_sp_list(Vars, SVars, SAdd1),
	generate_sp_list([Min, Max], [SMin, SMax], SAdd2),
	(
	  integer(SMin)
	->
	  VMin = SMin,
	  SAdd3 = SAdd1
	;
	  fresh_local_var_name(VMin),
	  append(SAdd1, [VMin #= SMin], SAdd3)
	),
	(
	  integer(SMax)
	->
	  VMax = SMax,
	  SAdd4 = SAdd3
	;	  
	  fresh_local_var_name(VMax),
	  append(SAdd3, [VMax #= SMax], SAdd4)
	),
	append(SAdd4, SAdd2, SAdd).

% fdvar. domain lower bound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_min(Var), FDMin, [fd_min(SVar, FDMin)]) :-
	!,
	generate_sp_code(Var, SVar, _).

% fdvar. domain upper bound
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_max(Var), FDMax, [fd_max(SVar, FDMax)]) :-
	!,
	generate_sp_code(Var, SVar, _).

% fdvar. domain cardinality
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(domain_size(Var), FDMax, [fd_size(SVar, FDMax)]) :-
	!,
	generate_sp_code(Var, SVar, _).

% global constraint disjoint2
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(disjoint2(Xs, Ys, Ss), disjoint2(Data), SAdd) :-
	!,
	generate_sp_code([Xs, Ys, Ss], [GXs, GYs, GSs], SAdd),
	disjoint_data(GXs, GYs, GSs, Data).

% global constraint cumulative
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(cumulative(Os, Ds, Hs, L),
		 cumulative(Tasks, [limit(SL), global(true)]), SAdd) :-
	!,
	generate_sp_code([Os, Ds, Hs], [SOs, SDs, SHs], _),
	set_cg_mode(dynamic),
	generate_sp_code(L, SL, SAdd1),
	set_cg_mode(static),	
	length(Os, N),
	gen_variables(N, Vs),
	mktasks(SOs, SDs, SHs, Vs, Tasks),
	append(SAdd1, [domain(Vs, 0, 1000)], SAdd).

% global constraint geost with rules
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(geost_constraint(PreObjects, PreSboxes, Dims),
		 rcp_geost(Objects, Sboxes, [non_overlapping(objects(Oids), D)]),
		 SAdd) :-
	!,
	generate_sp_geosr_object(PreObjects, ObjectsX, Oids, ObjectsShapeDoms),	
	generate_sp_code(ObjectsX, Objects, _),
	generate_sp_geosr_sbox(PreSboxes, Sboxs),
	generate_sp_list(ObjectsShapeDoms, SAdd, _),
	gen_integers(1,Dims, D),
	remove_dups(Sboxs, Sboxes).

% global constraint geost with rules + greedy mode options
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(geost_constraint(PreObjects, PreSboxes, Dims, Options),
		 rcp_geost(Objects, Sboxes, SOptions, [non_overlapping(objects(Oids), D)]),
		 SAdd) :-
	!,
	generate_sp_geosr_object(PreObjects, ObjectsX, Oids, ObjectsShapeDoms),	
	generate_sp_code(ObjectsX, Objects, _),
	generate_sp_geosr_sbox(PreSboxes, Sboxs),
	generate_sp_list(ObjectsShapeDoms, SAdd1, _),
	generate_sp_code(Options, SOptions, SAdd2),
	append(SAdd1, SAdd2, SAdd),	
	gen_integers(1,Dims, D),
	remove_dups(Sboxs, Sboxes).

generate_sp_code(object(SidSpec, OriginsSpec),
		 object('_', SSidSpec, SOriginsSpec), SAdd) :-
	generate_sp_list([SidSpec, OriginsSpec], [SSidSpec, SOriginsSpec], SAdd).
		 
% global constraint lexicographic
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(lexicographic(ListOfListOfVars),
		 lex_chain(SListOfListOfVars), SAdd) :-
	!,
	generate_sp_code(ListOfListOfVars, SListOfListOfVars, SAdd).

generate_sp_code(lexicographic_strict(ListOfListOfVars),
		 lex_chain(SListOfListOfVars, [op(#<)]), SAdd) :-
	!,
	generate_sp_code(ListOfListOfVars, SListOfListOfVars, SAdd).

% gui terms
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(gui_name(Var, AlphaNumList), 
		 gui_varnames([Var], [Atom]), []) :-
	!,
	atom_codes('_', SepCode),
	alphanum_concat(AlphaNumList, SepCode, NameCodeList),
	CL1 = [39 | NameCodeList],
	append(CL1,[39], CL),
	atom_codes(Atom, CL).

generate_sp_code(gui_control(button, Name, Term), 
		 gui_button(BTerm, STerm), SAdd) :-
	object_language(clpfdguisp),
	generate_sp_code(Term, STerm, SAdd),
	!,
	variables_set(STerm, Variables),
	BTerm =..[Name | Variables].

% atomic 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, []) :-
	atomic(Term),
	dict_sp(Term, STerm),
	!.
generate_sp_code(Term, Term, []) :-
	atomic(Term),
	!.

% list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	is_list(Term),
	!,
	generate_sp_list(Term, STerm, SAdd).

% arithmetic 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	is_arithop(Term),
	Term =.. [Op | Args],
	contains_reifop(Args),
	!,
        reify(Args, RArgs, SAdd),
	ST =.. [Op | RArgs],
	eval(ST, STerm).

% search (label) node
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(node(_L, Term), STerm, SAdd) :-
	!,
	generate_sp_code(Term, STerm, SAdd).

% user output 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(rcp_write(Term), rcp_write(STerm), SAdd) :-	
	!,
	generate_sp_code(Term, STerm, SAdd),
	format("\"--- User Output ---\"\n ~w ", [Term]),
	format("\n \"~~>\"\n",[]),
	format(" ~w\n\"--- End User Output ---\"\n", [STerm]).

% =
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(A = B, STerm, SAdd) :-
	!,
	generate_sp_code(A, SA, SAddA),
	generate_sp_code(B, SB, SAddB),
	append(SAddA, SAddB, SAdd),	
	(
	  is_builtin_expr(and(A, B))
	->
	  dict_sp(=, TOp),
	  ST =.. [TOp, SA, SB],
	  eval(ST, STerm)
	;
	  STerm =.. [=, SA, SB]
	).

% equiv
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(equiv(A, B), STerm, SAdd) :-
	!,
	generate_sp_code(A, SA, SAddA),
	generate_sp_code(B, SB, SAddB),
	append(SAddA, SAddB, SAdd),
	(
	  is_builtin_expr(and(A, B))
	->	    
	  dict_sp(equiv, TOp),
	  ST =.. [TOp, SA, SB],
	  eval(ST, STerm)
	;
	  STerm =.. [=, SA, SB]
	).

% default 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_code(Term, STerm, SAdd) :-
	Term =.. [Op | Args],
	generate_sp_list(Args, SArgs, SAdd),
	dict_sp(Op, TOp),
	ST =.. [TOp | SArgs],
	eval(ST, STerm).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% datastructure-dedicated code generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_list([], [], []).
generate_sp_list([UT | UTerms], [ST | STerms], SAdd) :-
	generate_sp_code(UT, ST, SA),
	generate_sp_list(UTerms, STerms, SAs),
	append(SA, SAs, SAdd).

% record 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_record([], [], []).
generate_sp_record([Key = Val | T], [Key = SVal | ST], SAdd) :-
	generate_sp_code(Val, SVal, SA),
	generate_sp_record(T, ST, SAs),
	append(SA, SAs, SAdd).

% range list 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_range_list([L .. U], (Lb .. Ub), SAdd) :-
	cg_mode(dynamic),
	!,
	generate_sp_code(L, Lb, SLb),
	generate_sp_code(U, Ub, SUb),
	append(SLb, SUb, SAdd).
generate_sp_range_list(L, L, []) :-
	cg_mode(dynamic),
	!.

generate_sp_range_list([L .. U], (L .. U), []) :-
	integer(L),
	integer(U),
	!.
generate_sp_range_list([L .. U], (inf .. U), []) :-
	L == min_integer,
	integer(U),
	!.
generate_sp_range_list([L .. U], (L .. sup), []) :-
	integer(L),
	U == max_integer,
	!.
generate_sp_range_list([S], {S}, []) :- 
	!,
	integer(S).
generate_sp_range_list([L .. U | T], RUnion \/ (L..U), SAdd) :-
	!,
	integer(L),
	integer(U),
	generate_sp_range_list(T, RUnion, SAdd).
generate_sp_range_list([S | T], RUnion \/ {S}, SAdd) :-
	integer(S),
	generate_sp_range_list(T, RUnion, SAdd).

% geosr  
%%%%%%%%%%%%%%%%%%%%
generate_sp_geosr_object([], [], [], []).
generate_sp_geosr_object([PreObj | PreObjs], [Obj | Objs], [Oid | Oids], ObjShpDoms) :-
	generate_sp_geosr_object(PreObjs, Objs, Oids, OSDoms),	
	PreObj = [Oid, S, Sids, Origins],	
	(
	 (
	   is_var(S)
	 ;
	   var(S)
	 )
	->
	  fresh_symbol('S', Sid),
	  append([Sid in Sids, element(S, Sids, Sid)], OSDoms, ObjShpDoms)
	;
	  nth1(S, Sids, Sid),
	  ObjShpDoms = OSDoms
	),
	Obj =.. [object, Oid, Sid, Origins, []].

generate_sp_geosr_sbox([], []).
generate_sp_geosr_sbox([PreSbox | PreSboxes], Sboxes) :-
	generate_sp_geosr_sbox(PreSboxes, Sboxs),
	traverse_presbox_shapes(PreSbox, Sbox),
	append(Sbox, Sboxs, Sboxes).

% explicit reification 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
reify(Term, RTerm, Add) :- 
	compound(Term),	
        (  
           is_compop(Term)
        ;
           is_logop(Term)
        ),
	!,
	Term =.. [Op | Args],
	reify(Args, RArgs, As),
	dict_sp(Op, TOp),
	ReifTerm =.. [TOp | RArgs],	
	fresh_symbol('R', ReifVar),
	RTerm = ReifVar,
	dict_sp(equiv, EquivOp),
	AddTerm =.. [EquivOp, ReifVar, ReifTerm],
	append(As, [AddTerm], Add).
reify([], [], []) :-
	!.
reify([H | T], [RT | RTs], Add) :-	
	reify(H, RT, A),
	reify(T, RTs, As),
	!,
	append(A, As, Add).
reify(Term, RTerm, Add) :-	
	generate_sp_code(Term, RTerm, Add).

contains_reifop([H | _]) :-
	contains_reifop(H),
	!.
contains_reifop([_ | T]) :-
	contains_reifop(T),
	!.
contains_reifop(Term):-
	compound(Term),
	is_compop(Term),
        !.
contains_reifop(Term):-
	compound(Term),
	is_logop(Term).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% statement object code generation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% foldr instance
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_statement(Head, Body-RCall, foldr, HeadCode-BodyCode) :-
	!,
% 				write('----------'),nl,
%  				write(Head),nl,
% 				write('-->'),nl,
%  				write_list(Body),nl,nl,

	Head =.. [FHead | ArgsHead],
	ArgsHead = [cons(_, var('Tail',0)), list(_Vs), Op, _Id],	
	generate_sp_list(ArgsHead, ArgsHeadCode0, _),
	ArgsHeadCode0 = [[Elem | _ElemTail], _Vars, _Operator, Identity],	
	ArgsHeadCode = [[Elem | _ElemTail], _Vars],
	
	(
	  srch_mode(_)
	->
	  atomize([rcp, FHead, search], '_', SFHead)
	;
	  atomize([rcp, FHead], '_', SFHead),
	  partition_term([Body])
	),
	
	cstr_terms(Cstrs),
	list2conj(Cstrs, Constraint),
	CompTerm =.. [Op, Constraint, RCall],
	generate_sp_code(CompTerm, CompTermCode, CompTermAddCode),
% 				write('--- Constraints'),nl,
% 				write_list(Constraints),nl,
% 				write('--- Compterm'),nl,
% 				write(CompTerm),nl,				
% 				write('--- CompTermAddCode'),nl,
% 				write(CompTermAddCode),nl,
% 				write('--- CompTermCode'),nl,
% 				write(CompTermCode),nl,					
	(
	  (
	    is_var_name(CompTermCode)
	  ;
	    integer(CompTermCode)
	  ;
	    is_list(CompTermCode)
	  ;
	    CompTermCode = [_ | _]
	  )	
	->
	  append(ArgsHeadCode, [CompTermCode], SArgsHeadCode),
	  HeadCode =.. [SFHead | SArgsHeadCode],	  
	  list2seq(CompTermAddCode, BodyCode)
	;	  
	  append(ArgsHeadCode, ['_'], SArgsHeadCode),
 	  HeadCode =.. [SFHead | SArgsHeadCode],	  
	  append(CompTermAddCode, [CompTermCode], BodyCode0),
	  list2seq(BodyCode0, BodyCode)
	),
					  
	TerminalHead =.. [SFHead, [], '_', Identity],
	TerminalBody = '!',
	assert(statement(TerminalHead, TerminalBody, foldr)).	

generate_sp_statement(Head, Body-RCall, foldl, HeadCode-BodyCode) :-
	!,
% 				write('----------'),nl,
%  				write(Head),nl,
% 				write('-->'),nl,
%  				write_list(Body),nl,nl,
				
	Head =.. [FHead | ArgsHead0],
	ArgsHead0 = [cons(H, var('Tail',0)), list(_Vs), Op],
	ArgsHead = [cons(H, var('Tail',0)), var('Identity', 0), list(_Vs)],
	generate_sp_list(ArgsHead, ArgsHeadCode, _),
	ArgsHeadCode = [[_Elem | _ElemTail], Identity, _Vars],

	(
	  srch_mode(_)
	->
	  atomize([rcp, FHead, search], '_', SFHead)
	;
	  atomize([rcp, FHead], '_', SFHead),
	  partition_term([Body])
	),
	
	cstr_terms(Cstrs),
	list2conj(Cstrs, Constraint),
	composition(Op, Identity, Constraint, CompTerm),	
	generate_sp_code(CompTerm, CompTermCode, CompTermAddCode),
% 				write('--- Constraints'),nl,
% 				write_list(Constraint),nl,
% 				write('--- Compterm'),nl,
% 				write(CompTerm),nl,				
% 				write('--- CompTermAddCode'),nl,
% 				write(CompTermAddCode),nl,
% 				write('--- CompTermCode'),nl,
% 				write(CompTermCode),nl,		
	(
	  (
	    is_var_name(CompTermCode)
	  ;
	    integer(CompTermCode)
	  ;
	    is_list(CompTermCode)
	  ;
	    CompTermCode = [_ | _]
	  )	
	->
	  R = CompTermCode,
	
	  CompTermAddCode = BodyCode0
	;	  

	  fresh_local_var_name(R),
	  
	  append(CompTermAddCode, [CompTermCode], BodyCode0)	  
	),
% 				write('--- R'),nl,
% 				write(R),nl,
% 				write('--- BodyCode1'),nl,
% 				write(BodyCode0),nl,
	generate_sp_code(RCall, RCallCode, RCallAddCode),
% 				write('--- RCallCode'),nl,
% 				write(RCallCode),nl,
% 				write('--- RCallAddCode'),nl,
% 				write(RCallAddCode),nl,
	last(RCallAddCode, RecCode0),
	delete(RCallAddCode, RecCode0, SAddRecCall1),	
	RecCode0 =.. [RCallFunctor, Tail, Env, _RCallCode],
	RecCode =.. [RCallFunctor, Tail, R, Env, _RCallCode],				
 	append(SAddRecCall1, [RecCode], RecCallCode),
	append(BodyCode0, RecCallCode, BodyCode),

	append(ArgsHeadCode, [RCallCode], SArgsHeadCode),
	HeadCode =.. [SFHead | SArgsHeadCode],
	
	TerminalHead =.. [SFHead, [], Identity, '_', Identity],
	TerminalBody = '!',
	assert(statement(TerminalHead, TerminalBody, foldl)).	

generate_sp_statement(Head, _Body, _T, none-none) :-
	srch_mode(_),
	functor(Head, Functor, Arity),
	lang(dynvordercrit_preds, VOrderPreds),	
	memberchk(Functor/Arity, VOrderPreds),
	!.

% definition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
generate_sp_statement(Head, Body, _T, SHead-SBody) :-
	!,
  				%write('----------'),nl,
  				%write(Head),nl,
  				%write(Body),nl,
  				%write('----------'),nl,
	Head =.. [FHead | HeadArgs],
	generate_sp_list(HeadArgs, SHeadArgs, _SA),
	SHead1 =.. [FHead | SHeadArgs],
	counter_init(gen_local_var_names),	
	(
	  srch_mode(_)
	->	  
	  atomize([rcp, FHead, search], '_', SFHead)	
	;
	  subst_free_var(Body, SHead1, BodySFV),
	  partition_term([BodySFV]),
	  atomize([rcp, FHead], '_', SFHead)	
	),
	cstr_terms(Constraints),
	generate_sp_code(list(Constraints), SBody1, SAdd1),
				%write(body(SBody1)),nl,
 				%write(add(SAdd1)),nl,nl,
	
	delete_variables(SBody1, SBodyWOV, LV),
 	(
	  LV == novar
	->
	  (
	    SBody1 = [Sbody1Content],
	    is_list(Sbody1Content)
	  ->
	    fresh_local_var_name(RArgHead),
	    delete(SBodyWOV, Sbody1Content, SBodyWOV0),
	    append(SBodyWOV0, [RArgHead = Sbody1Content], AppSBody1),
	    append(SAdd1, AppSBody1, SBody)
	  ;
	    fresh_local_var_name(RArgHead),
	    append(SBodyWOV, [fd_aeq(RArgHead, 1)], AppSBody1),
	    append(SAdd1, AppSBody1, SBody)
	  )
	;
	  RArgHead = LV,
	  append(SAdd1, SBodyWOV, SBody)
	),	
	append(SHeadArgs, [RArgHead], RHeadArgs),
	SHead =.. [SFHead | RHeadArgs].

generate_sp_statement(Head, Body, T, _-_) :-
	Term =.. [T, Head, Body],
	object_language(ObjLang),
	raise_exception(code_generation(not_accepted_by_language(ObjLang),
					 [Term], [])).

generate_sp_statements :-
	findall(statement(Head, Body, T),
		statement(Head, Body, T),
		LStatements),
	cg_mode(CGModeBefore),
	generate_sp_statements(LStatements),
	set_cg_mode(CGModeBefore).

generate_sp_statements([]) :-
	!.
generate_sp_statements([statement(Head, Body, T) | LDynClauses]) :-
				 %write(statement(Head, Body, T)),nl,
	init_objcodegenerator,
	set_cg_mode(dynamic),
	retract(statement(Head, Body, T)),
				 %write('deterministic version'),nl,
	generate_sp_statement(Head, Body, T, SHead-SBody),
	assert_statement(SHead, SBody, T),
				 %write('non-deterministic version'),nl,
 	set_srch_mode(direct), 	
 	generate_sp_statement(Head, Body, T, SHeadSrch-SBodySrch),	
	assert_statement(SHeadSrch, SBodySrch, T),
	unset_srch_mode,	
	generate_sp_statements(LDynClauses).

assert_statement(none, none, _Type) :-
	!.
assert_statement(Head, Body, Type) :-
	assert(statement(Head, Body, Type)).

