% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file ocw.pl
%

:- ensure_loaded([ocg]).
:- ensure_loaded([misc]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% object-language-specific code writing 
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% clp(fd) SICStus
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_code(CodeList, Files) :- 
	object_language(clpfdsp),
	!,
	last(Files, FullInFileName),
	atom_concat(InFileName, '.rcp', FullInFileName),
	atom_concat(InFileName, '.pl', OutFileName),
	split_string(OutFileName, '/', StrSplitted),
	last(StrSplitted, File0),
	atom_concat(['./', File0], File),
	format("Writing file \"~w\"... \n", [File]),	
	open(File, write, Stream),	
	write_imports(Stream),
	%write_rcp_statements(Stream),
	write_statements(Stream),	
	write_query(CodeList, Stream),	
	format("File \"~w\" wrote.\n", [File]),
	close(Stream).

% Choco JAVA
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_code(CodeList, Files) :- 
	object_language(clpfdcj),
	!,
	last(Files, FullInFileName),
	atom_concat(InFileName, '.rcp', FullInFileName),
	atom_concat(InFileName, '.java', OutFileName),
	split_string(OutFileName, '/', StrSplitted),
	last(StrSplitted, File0),
	atom_concat(['./', File0], File),
	format("Writing file \"~w\"... \n", [File]),
	open(File, write, Stream),
	split_string(InFileName, '/', InFileNameSplitted),
	last(InFileNameSplitted, Name),
	write_package(Stream, InFileName),
	write_imports(Stream),
				%write_statements(Stream),
	write_query([Name | CodeList], Stream),	
	format("File \"~w\" wrote.\n", [File]),
	close(Stream).

% clpfd + clpgui SICStus
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_code(LowCodeList, Files) :-
	object_language(clpfdguisp),	
	last(Files, FullInFileName),
	atom_concat(InFileName, '.rcp', FullInFileName),
	atom_concat(InFileName, '.pl', File),	
	open(File, write, Stream),
	format("Writing file \"~w\"... \n",[File]),
	write_imports(Stream),	
	LowCodeList = [GuiVarNames,GuiControls,Constraints],
	split(FullInFileName, '/', InFileNameList, []),
	last(InFileNameList, FileName),	
	atom_concat(FileNamePrefix, '.rcp', FileName),	
	write(Stream, FileNamePrefix),
	write(Stream, ' :- \n\t'),	
	write_list(Stream, ',\n\t', GuiVarNames),
	(GuiVarNames == [] -> true; write(Stream, ',\n\t')),		
	write(Stream, 'gui_show_values'),
	(Constraints == [] -> true; write(Stream, ',\n\t')),
	write_list(Stream, ',\n\t', Constraints),
	write_control_calls(Stream, constraints, GuiControls),
	write_control_calls(Stream, search, GuiControls),
	write(Stream, '.'),	
	write_control_clauses(Stream, GuiControls),	
	format("File \"~w\" wrote.\n",[File]),
	close(Stream).

% write terms
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
my_write_term(Stream, Term) :-
	var(Term),
	!,
	write_term(Stream, Term, [portrayed(true), indented(true)]).

% Choco Java
%%%%%%%%%%%%%%%%%%%%
my_write_term(Stream, cj_string(Term)) :-
	object_language(clpfdcj),
	!,	
	write(Stream, '\"'),
	my_write_term(Stream, Term),
	write(Stream, '\"').
my_write_term(Stream, cj_generic(Generic, Type)) :-
	object_language(clpfdcj),
	!,
	to_upper_case(Generic, G),	
	write(Stream, G),
	write(Stream, '<'),
	to_upper_case(Type, T),
	my_write_term(Stream, T),
	write(Stream, '>').
my_write_term(Stream, cj_var_decl(Type, TermL) = TermR) :-
	object_language(clpfdcj),
	!,
	my_write_term(Stream, Type),
	write(Stream, ' '),
	my_write_term(Stream, TermL),
	write(Stream, ' = '),
	my_write_term(Stream, TermR).
my_write_term(Stream, cj_obj_new(cj_generic(G, T))) :-
	object_language(clpfdcj),
	!,	
	write(Stream, 'new '),
	my_write_term(Stream, cj_generic(G, T)),
	write(Stream, '()').
my_write_term(Stream, cj_obj_new(Term)) :-
	object_language(clpfdcj),
	atom(Term),
	!,	
	write(Stream, 'new '),
	my_write_term(Stream, cj_class(Term)),
	write(Stream, '()').
my_write_term(Stream, cj_obj_new(Term)) :-
	object_language(clpfdcj),
	!,	
	write(Stream, 'new '),
	my_write_term(Stream, cj_class(Term)).
my_write_term(Stream, cj_obj_new(cj_array_of(Type), Term)) :-
	object_language(clpfdcj),
	!,	
	write(Stream, 'new '),
	my_write_term(Stream, cj_class(Type)),
	write(Stream, '[] '),
	write(Stream, '{'),
	write_list(Stream, ', ', Term),
	write(Stream, '}').
my_write_term(Stream, cj_pkg_path(Path)) :-
	object_language(clpfdcj),
	!,
	atomize(Path, '.', APath),
	write(Stream, APath).
my_write_term(Stream, cj_array_of(Type)) :-
	object_language(clpfdcj),
	!,
	my_write_term(Stream, Type),
	write(Stream, '[] ').
my_write_term(Stream, cj_obj_field(Obj, Field)) :-
	object_language(clpfdcj),
	!,
	my_write_term(Stream, Obj),
	write(Stream, '.'),
	my_write_term(Stream, Field).
my_write_term(Stream, cj_obj_mthd_call(Obj, MthdCall)) :-
	object_language(clpfdcj),
	atom(MthdCall),
	!,
	write(Stream, Obj),
	write(Stream, '.'),
	my_write_term(Stream, MthdCall),
	write(Stream, '()').
my_write_term(Stream, cj_obj_mthd_call(Obj, MthdCall)) :-
	object_language(clpfdcj),
	!,
	write(Stream, Obj),
	write(Stream, '.'),
	my_write_term(Stream, MthdCall).
my_write_term(Stream, cj_obj_static_mthd_call(Obj, MthdCall)) :-
	object_language(clpfdcj),
	atom(MthdCall),
	!,
	to_upper_case(Obj, ObjUppCase),
	write(Stream, ObjUppCase),
	write(Stream, '.'),
	my_write_term(Stream, MthdCall),
	write(Stream, '()').
my_write_term(Stream, cj_obj_static_mthd_call(Obj, MthdCall)) :-
	object_language(clpfdcj),
	!,
	to_upper_case(Obj, ObjUppCase),
	write(Stream, ObjUppCase),
	write(Stream, '.'),
	my_write_term(Stream, MthdCall).
my_write_term(Stream, cj_class(Term)) :-
	object_language(clpfdcj),
	memberchk(Term, [int, string, bool]),
	!,
	write(Stream, Term).
my_write_term(Stream, cj_class(Term)) :-
	object_language(clpfdcj),
	atomic(Term),
	!,
	to_upper_case(Term, T),
	write(Stream, T).
my_write_term(Stream, cj_class(Term)) :-
	object_language(clpfdcj),
	compound(Term),
	!,
	Term =.. [F | Args],
	to_upper_case(F, Functor),
	write(Stream, Functor),
	write(Stream, '('),
	my_write_term_args(Stream, Args),
	write(Stream, ')').
my_write_term(Stream, Term) :-
	object_language(clpfdcj),
	compound(Term),
	!,
	Term =.. [Functor | Args],
	write(Stream, Functor),
	write(Stream, '('),
	my_write_term_args(Stream, Args),
	write(Stream, ')').

% SICStus Prolog
my_write_term(Stream, [H | T ]) :-
	object_language(clpfdsp),
	is_var_name(T),
	!,
	write(Stream, '['),
	my_write_term(Stream, H),
	write(Stream, '|'),
	my_write_term(Stream, T),
	write(Stream, ']').

my_write_term(Stream, Term) :-
	object_language(clpfdsp),
	is_list(Term),
	!,
	write(Stream, '['),
	write_list(Stream, ',', Term),
	write(Stream, ']').

% my_write_term(Stream, Term) :-
% 	object_language(clpfdsp),
% %	icg_mode,
% 	!,
% 	write_term(Stream, Term, [portrayed(true), indented(true), quoted(true)]).

% default
%%%%%%%%%%%%%%%%%%%%
my_write_term(Stream, Term) :-
	write_term(Stream, Term, [portrayed(true), indented(true)]).

% compoud term arguments
my_write_term_args(_, []).
my_write_term_args(Stream, [Arg]) :-
	!,
	my_write_term(Stream, Arg).
my_write_term_args(Stream, [Arg | Args]) :-
	my_write_term(Stream, Arg),
	write(Stream, ', '),
	my_write_term_args(Stream, Args).

% write list of terms
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_list(_, _, []).
write_list(Stream, Sep, [H]) :-
	object_language(clpfdcj),
	Sep == ';\n    ',
	!,
	write_sequence(Stream, Sep, H),
	write(Stream, Sep).
write_list(Stream, Sep, [H]) :-
	!,
	write_sequence(Stream, Sep, H).
write_list(Stream, Sep, [H | T]) :-
	write_sequence(Stream, Sep, H),
	my_write_term(Stream, Sep),
	write_list(Stream, Sep, T).

% write sequence of terms
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_sequence(Stream, Sep, ','(true, T)) :-
	!,
	write_sequence(Stream, Sep, T).
write_sequence(Stream, Sep, ','(H, true)) :-
	!,
	write_sequence(Stream, Sep, H).
write_sequence(Stream, Sep, ','(H, T)) :-
	!,
	write_sequence(Stream, Sep, H),
	my_write_term(Stream, Sep),
	write_sequence(Stream, Sep, T).
write_sequence(Stream, Sep, ';'(H, T)) :-
	object_language(clpfdsp),
	!,
	write(Stream, '(\n\t'),	
	write_sequence(Stream, Sep, H),
	write(Stream, '\n\t'),	
	my_write_term(Stream, ';'),
	write(Stream, '\n\t'),	
	write_sequence(Stream, Sep, T),
	write(Stream, '\n\t)').
write_sequence(Stream, _, H) :-
	object_language(clpfdcj),
	functor(H, F, A),
	(
	  F == reifiedIntConstraint
	;
	  is_gcstr(F/A)
	;
	  dict_cj(Op, F),
	  functor(T, Op, A),
	  (
	    is_compop(T)
	  ;
	    is_logop(T)
	  )
	),
	!,
	write(Stream, 'addConstraint('),
	my_write_term(Stream, H),
	write(Stream, ')').
write_sequence(Stream, _, H) :-
	my_write_term(Stream, H).

% write query
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_query([DecisionVariables, ConstraintsList], Stream) :-
	object_language(clpfdsp),
	!,
	QueryHead =.. [solve | [DecisionVariables]],
	implant_statistics_code(ConstraintsList, QueryBody),
	write_clause(Stream, QueryHead, QueryBody).
write_query([FileName, _DecisionVariables, ConstraintsList], Stream) :-
	object_language(clpfdcj),
	!,
	write_class(Stream, FileName, ConstraintsList).
write_query(_, _).

implant_statistics_code(CodeList, ImplantedStatCodeList) :-
	object_language(clpfdsp),
	!,
	CodeList0 = [statistics(runtime, ['_Runtime_0_'| '_']),
		     fd_statistics(backtracks, '_') | CodeList],
	append(CodeList0,
	       [statistics(runtime, ['_Runtime_1_' | '_']),
		fd_statistics(backtracks, '_Backtracks_'),	
		'_Runtime_' is '_Runtime_1_' - '_Runtime_0_',
		format('\'~3d seconds\\n\'', ['_Runtime_']),
		format('\'~w bactracks\\n\\n\'', ['_Backtracks_'])],
	       ImplantedStatCodeList).
implant_statistics_code(CodeList, CodeList).

% write module imports
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_imports(Stream) :-
	object_language(clpfdsp),
	!,
	rcp_dir(Root),
	atom_concat(['\'', Root, '/rules2cp/interpreter/prolog/r2cp.pl', '\''],
		    File),
	write_imports_rec(Stream, [use_module(library(clpfd)),
				   use_module(File),
				   init_interpreter]).
write_imports(Stream) :-
	object_language(clpfdcj),
	!,
	I = ['static rules2cp.interpreter.java.r2cp.*',
	     'rules2cp.interpreter.java.*',
	     'java.util.Vector',
	     'static choco.Choco.*',
	     'choco.cp.model.CPModel',
	     'choco.cp.solver.CPSolver',
	     'choco.cp.solver.search.integer.varselector.*',
	     'choco.cp.solver.search.integer.valselector.*',
	     'choco.cp.solver.search.integer.valiterator.*',
	     'choco.cp.solver.search.integer.branching.*',
	     'choco.cp.solver.constraints.integer.*',
	     'choco.cp.solver.constraints.global.geost.Constants',
	     'choco.cp.solver.constraints.global.geost.externalConstraints.ExternalConstraint',
	     'choco.cp.solver.constraints.global.geost.externalConstraints.NonOverlapping',
	     'choco.cp.solver.constraints.global.geost.geometricPrim.ShiftedBox',
	     'choco.kernel.model.Model',
	     'choco.kernel.model.constraints.*',
	     'choco.kernel.model.variables.*',
	     'choco.kernel.model.variables.integer.*',				   
	     'choco.kernel.model.variables.geost.GeostObject',				   
	     'choco.kernel.solver.Solver',
	     'choco.kernel.solver.branch.*',
	     'choco.kernel.solver.variables.integer.*',
	     'choco.kernel.model.variables.geost.*'],
	write_imports_rec(Stream, I).
write_imports(Stream) :-
	write_imports_rec(Stream, []).

write_imports_rec(Stream, []) :-
	!,
	write(Stream, '\n').
write_imports_rec(Stream, [ImportTerm | ImportTerms]) :-
	object_language_class(clpfd_cj),
	!,
	write(Stream, 'import '),
	my_write_term(Stream, ImportTerm),
	write(Stream, ';\n'),
	write_imports_rec(Stream, ImportTerms).
write_imports_rec(Stream, [ImportTerm | ImportTerms]) :-
	object_language_class(clpfd_sp),
	!,
	write(Stream, ':- '),
	my_write_term(Stream, ImportTerm),
	write(Stream, '.\n'),
	write_imports_rec(Stream, ImportTerms).
write_imports_rec(_, _).

% write package  
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_package(Stream, Name) :-
	object_language(clpfdcj),
	split_string(Name, '/', NameSplitted),
	sublist(NameSplitted, PName, 0, _, 1),
	atomize(PName, '.', PackageName),
	write(Stream, 'package '),
	write(Stream, PackageName),
	write(Stream, ';\n').

% write class
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_class(_, _, []).
write_class(Stream, ClassName, BodyAsList) :-
	object_language(clpfdcj),
	write(Stream, 'public class '),
	write(Stream, ClassName),
	write(Stream, ' {\n\n  '),
	write(Stream, 'public static void main(String[] args) {\n\n    '),
	write_list(Stream, ';\n    ', BodyAsList),
	write(Stream, '\n  '),
	write(Stream, '}\n\n'),
	write(Stream, '}\n').

% write clause
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_clause(_, _, []).
write_clause(Stream, Head, BodyAsList) :-
	is_list(BodyAsList),
	!,
	my_write_term(Stream, Head),
	write(Stream, ' :- \n\t'),
	write_list(Stream, ',\n\t', BodyAsList),
	write(Stream, '.\n').
write_clause(Stream, Head, Body) :-
	my_write_term(Stream, Head),
	write(Stream, ' :- \n\t'),
	write_sequence(Stream, ',\n\t', Body),
	write(Stream, '.\n').

% fact
write_fact(Stream, Term) :-
	my_write_term(Stream, Term),
	write(Stream, '.\n').

% write statement
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% interpreted statement
write_rcp_statements(Stream) :-
	findall(rcp_statement(Head, Body, T),
		rcp_statement(Head, Body, T), LDynClauses),
	cg_mode(CGModeBefore),
	set_cg_mode(dynamic),
        set_icg_mode,
        write_rcp_statements(Stream, LDynClauses),
        unset_icg_mode,
	set_cg_mode(CGModeBefore).
	
write_rcp_statements(Stream, []) :-
	!,
	write(Stream, '\n').
write_rcp_statements(Stream, [rcp_statement(Head, Body, T) | LDynClauses]) :-
	Fact = rcp_rule(Head, Body),
	write_fact(Stream, Fact),
	retract(rcp_statement(Head, Body, T)),
	write_rcp_statements(Stream, LDynClauses).


% compiled statement
write_statements(Stream) :-
	findall( statement(Head, Body, T),
		 statement(Head, Body, T), LDynClauses),	
	cg_mode(CGModeBefore),
	write_statements(Stream, none, LDynClauses),	
	set_cg_mode(CGModeBefore).

write_statements(Stream, _, []) :-
	!,
	write(Stream, '\n').
write_statements(Stream, LastPred, [statement(Head, Body, T) | LDynClauses]) :-
	functor(Head, F, A),
	(
	  F/A == LastPred
	->
	  true
	;
	  write(Stream, '\n')
	),
	write_clause(Stream, Head, Body),
	retract(statement(Head, Body, T)),
	write_statements(Stream, F/A, LDynClauses).

% write clpgui control clause
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
write_control_calls(_,_, []).
write_control_calls(Stream, constraints, [GuiControl | GuiControls]) :- 
	object_language(clpfdguisp), 
	GuiControl = gui_button(Head, _),
	functor(Head, Functor, Arity),
	lang(search_preds, SearchPreds),
	\+ memberchk(Functor/Arity, SearchPreds),
	write(Stream, ',\n\t'),
	write(Stream, 'gui_button('),
	write(Stream, Head),
	write(Stream, ')'),
	write_control_calls(Stream, constraints, GuiControls).
write_control_calls(Stream, search, [GuiControl | GuiControls]) :- 
	object_language(clpfdguisp), 
	GuiControl = gui_button(Head, _),
	functor(Head, Functor, Arity),
	lang(search_preds, SearchPreds),
	memberchk(Functor/Arity, SearchPreds),
	write(Stream, ',\n\t'),
	write(Stream, 'gui_button('),
	write(Stream, Head),
	write(Stream, ')'),
	write_control_calls(Stream, search, GuiControls).
write_control_calls(Stream, Type, [_ | GuiControls]) :-
	write_control_calls(Stream, Type, GuiControls).

write_control_clauses(_, []).
write_control_clauses(Stream, [gui_button(Head, BodyAsList) | GuiControls]) :-
	object_language(clpfdguisp),
	write(Stream, '\n\n'),
	write(Stream, Head),
	write(Stream, ' :- \n\t'),
	write_list(Stream, ',\n\t', BodyAsList),
	write(Stream, '.'),
	write_control_clauses(Stream, GuiControls).

is_gcstr(allDifferent/1).
is_gcstr(cumulative/4).
is_gcstr(geost/4).
