% Copyright 2008 INRIA, F. Fages & J. Martin
% License GPL
%
% file r2cp.pl
%
:- ensure_loaded(library(system)).
:- ensure_loaded([err]).
:- ensure_loaded([compiler]).

% r2cp(['rules2cp/examples/nqueens/queens.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).
% r2cpchoco(['rules2cp/examples/nqueens/queens.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).

% r2cp(['rules2cp/examples/scheduling/scheduling.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).
% r2cpchoco(['rules2cp/examples/scheduling/scheduling.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).

% r2cp(['rules2cp/examples/scheduling/bridge.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).
% r2cpchoco(['rules2cp/examples/scheduling/bridge.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).

% r2cp(['rules2cp/examples/packing/packing.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).
% r2cpchoco(['rules2cp/examples/packing/packing.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).

% r2cp(['rules2cp/examples/packing/packing_polymorphism.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).
% r2cpchoco(['rules2cp/examples/packing/packing_polymorphism.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).

% r2cp(['rules2cp/examples/packing/psa.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).
% r2cpchoco(['rules2cp/examples/packing/psa.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).

% r2cp(['rules2cp/examples/packing/rpacking.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).
% r2cpchoco(['rules2cp/examples/packing/rpacking.rcp', '-rcpdir', '/afs/inria.fr/rocq/home/contraintes/jmartin']).

% r2cpgui(['rules2cp/examples/packing/psa_gui_pattern.rcp']).

:- dynamic rcp_path/1,
	   rcp_directory/1.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% clpfd - SICStus Prolog
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% command-line
%%%%%%%%%%%%%%%%%%%%%
r2cp :-
	prolog_flag(argv, Argv),
	on_exception(Error, r2cp(Argv), error_handler(Error)).

make_r2cp(RcpDir) :-
	save_program(r2cp, (set_rcp_dir(RcpDir), r2cp, halt)).

% top-level
%%%%%%%%%%%%%%%%%%%%%
r2cp(Argv) :-
	r2cp_arguments(Argv, Files),
	compile(Files, clpfdsp).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% clpfd - CHOCO Java
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% command-line
%%%%%%%%%%%%%%%%%%%%%
r2cpchoco :-
	prolog_flag(argv, Argv),
	on_exception(Error, r2cpchoco(Argv), error_handler(Error)).

make_r2cpchoco(RcpDir) :-
	save_program(r2cpchoco, (set_rcp_dir(RcpDir), r2cpchoco, halt)).

% top-level
%%%%%%%%%%%%%%%%%%%%%
r2cpchoco(Argv) :-
	r2cp_arguments(Argv, Files),
	compile(Files, clpfdcj).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% r2cp installation path
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
set_rcp_dir(Directory) :-
	nonvar(Directory), 
	retractall(rcp_directory(_)),
	assert(rcp_directory(Directory)).

rcp_dir(Directory) :-
	rcp_directory(Directory).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% r2cp sources path
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
set_rcp_path(Directory) :-
	nonvar(Directory), 
	retractall(rcp_path(_)),
	assert(rcp_path(Directory)).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% command-line arguments treatment
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
r2cp_arguments([], []).
r2cp_arguments(['-rcppath', Path | Argv], Files) :-
	!,
	set_rcp_path(Path),
	r2cp_arguments(Argv, Files).
r2cp_arguments(['-rcpdir', Path | Argv], Files) :-
	!,
	set_rcp_dir(Path),
	r2cp_arguments(Argv, Files).
r2cp_arguments([Arg | Argv], [Arg | Files]) :-
	r2cp_arguments(Argv, Files).
