:-use_module(library(lists)).
:-use_module(library(clpfd)).
:-ensure_loaded(geosr).

/* example for the paper */

example(S, U, V) :-
	S in 3..4, U in 1..9, V in 1..6,
	geosr([object(1,1,[1,2],[type-2]),
	       object(2,2,[3,3],[type-1]),
	       object(3,S,[2,5],[type-2]),
	       object(4,1,[3,7],[type-1]),
	       object(5,5,[U,V],[type-1])],
	      [sbox(1,[0,0],[3,1]),
	       sbox(2,[0,0],[1,1]),
	       sbox(3,[0,0],[1,2]),
	       sbox(4,[0,0],[2,1]),
	       sbox(5,[0,0],[2,2])],
	      [(origin(O1,S1,D) ---> O1^x(D)+S1^t(D)),
	       (end(O1,S1,D) ---> O1^x(D)+S1^t(D)+S1^l(D)),
	       (overlap_sboxes(Dims, O1, S1, O2, S2) --->
		   forall(D, Dims,
			  end(O1,S1,D) #> origin(O2,S2,D) #/\
			  end(O2,S2,D) #> origin(O1,S1,D))),
	       (meet_sboxes(Dims, O1, S1, O2, S2) --->
		   forall(D, Dims,
			  end(O1,S1,D) #>= origin(O2,S2,D) #/\
			  end(O2,S2,D) #>= origin(O1,S1,D)) #/\
	           exists(D, Dims,
			  end(O1,S1,D) #= origin(O2,S2,D) #\/ % N.B.
			  end(O2,S2,D) #= origin(O1,S1,D))), %  N.B.
	       (all_not_overlap_sboxes(Dims,OIDs) --->
		   forall(O1,objects(OIDs),
			  forall(S1,sboxes([O1^sid]),
				 forall(O2,objects(OIDs),
					O1^oid #< O2^oid #=>
				        forall(S2,sboxes([O2^sid]),
					       #\ overlap_sboxes(Dims,O1,S1,O2,S2)))))),
	       (all_type1_not_meet_sboxes(Dims,OIDs) --->
		   forall(O1,objects(OIDs),
			  forall(S1,sboxes([O1^sid]),
				 forall(O2,objects(OIDs),
					O1^oid #< O2^oid #/\ O1^type#=1 #/\ O2^type#=1 #=>
				        forall(S2,sboxes([O2^sid]),
					       #\ meet_sboxes(Dims,O1,S1,O2,S2)))))),
	       all_not_overlap_sboxes([1,2],[1,2,3,4,5]),
	       all_type1_not_meet_sboxes([1,2],[1,2,3,4,5])]).

example2(U, V) :-
	U in 1..9, V in 1..6,
	geosr([object(1,1,[1,2],[type-2]),
	       object(2,2,[3,3],[type-1]),
	       object(4,1,[3,7],[type-1]),
	       object(5,5,[U,V],[type-1])],
	      [sbox(1,[0,0],[3,1]),
	       sbox(2,[0,0],[1,1]),
	       sbox(4,[0,0],[2,1]),
	       sbox(5,[0,0],[2,2])],
	      [(origin(O1,S1,D) ---> O1^x(D)+S1^t(D)),
	       (end(O1,S1,D) ---> O1^x(D)+S1^t(D)+S1^l(D)),
	       (overlap_sboxes(Dims, O1, S1, O2, S2) --->
		   forall(D, Dims,
			  end(O1,S1,D) #> origin(O2,S2,D) #/\
			  end(O2,S2,D) #> origin(O1,S1,D))),
	       (meet_sboxes(Dims, O1, S1, O2, S2) --->
		   forall(D, Dims,
			  end(O1,S1,D) #>= origin(O2,S2,D) #/\
			  end(O2,S2,D) #>= origin(O1,S1,D)) #/\
	           exists(D, Dims,
			  end(O1,S1,D) #=< origin(O2,S2,D) #\/
			  end(O2,S2,D) #=< origin(O1,S1,D))),
	       (all_not_overlap_sboxes(Dims,OIDs) --->
		   forall(O1,objects(OIDs),
			  forall(S1,sboxes([O1^sid]),
				 forall(O2,objects(OIDs),
					O1^oid #< O2^oid #=>
				        forall(S2,sboxes([O2^sid]),
					       #\ overlap_sboxes(Dims,O1,S1,O2,S2)))))),
	       (all_type1_not_meet_sboxes(Dims,OIDs) --->
		   forall(O1,objects(OIDs),
			  forall(S1,sboxes([O1^sid]),
				 forall(O2,objects(OIDs),
					O1^oid #< O2^oid #/\ O1^type#=1 #/\ O2^type#=1 #=>
				        forall(S2,sboxes([O2^sid]),
					       #\ meet_sboxes(Dims,O1,S1,O2,S2)))))),
	       all_not_overlap_sboxes([1,2],[1,2,4,5]),
	       all_type1_not_meet_sboxes([1,2],[1,2,4,5])]).

end_of_file.

% after rewrite_rules

\bigwedge\left(
         \begin{array}{c}
         \neg\left(
             \begin{array}{c}
             \bigwedge\left(
                      \begin{array}{c}
                      4>x_{51} \\
                      x_{51}+2>1 \\
                      3>x_{52} \\
                      x_{52}+2>2 \\
                      \end{array}
                      \right) \\
             \end{array}
             \right) \\
         \neg\left(
             \begin{array}{c}
             \bigwedge\left(
                      \begin{array}{c}
                      4>x_{51} \\
                      x_{51}+2>3 \\
                      4>x_{52} \\
                      x_{52}+2>3 \\
                      \end{array}
                      \right) \\
             \end{array}
             \right) \\
         \neg\left(
             \begin{array}{c}
             \bigwedge\left(
                      \begin{array}{c}
                      2+\ell_{31}>3 \\
                      5+\ell_{32}>7 \\
                      \end{array}
                      \right) \\
             \end{array}
             \right) \\
         \neg\left(
             \begin{array}{c}
             \bigwedge\left(
                      \begin{array}{c}
                      2+\ell_{31}>x_{51} \\
                      x_{51}+2>2 \\
                      5+\ell_{32}>x_{52} \\
                      x_{52}+2>5 \\
                      \end{array}
                      \right) \\
             \end{array}
             \right) \\
         \neg\left(
             \begin{array}{c}
             \bigwedge\left(
                      \begin{array}{c}
                      6>x_{51} \\
                      x_{51}+2>3 \\
                      8>x_{52} \\
                      x_{52}+2>7 \\
                      \end{array}
                      \right) \\
             \end{array}
             \right) \\
         \neg\left(
             \begin{array}{c}
             \bigwedge\left(
                      \begin{array}{c}
                      4 \geq x_{51} \\
                      x_{51}+2 \geq 3 \\
                      4 \geq x_{52} \\
                      x_{52}+2 \geq 3 \\
                      \bigvee\left(
                             \begin{array}{c}
                             4=x_{51} \\
                             x_{51}+2=3 \\
                             4=x_{52} \\
                             x_{52}+2=3 \\
                             \end{array}
                             \right) \\
                      \end{array}
                      \right) \\
             \end{array}
             \right) \\
         \neg\left(
             \begin{array}{c}
             \bigwedge\left(
                      \begin{array}{c}
                      6 \geq x_{51} \\
                      x_{51}+2 \geq 3 \\
                      8 \geq x_{52} \\
                      x_{52}+2 \geq 7 \\
                      \bigvee\left(
                             \begin{array}{c}
                             6=x_{51} \\
                             x_{51}+2=3 \\
                             8=x_{52} \\
                             x_{52}+2=7 \\
                             \end{array}
                             \right) \\
                      \end{array}
                      \right) \\
             \end{array}
             \right) \\
         \end{array}
         \right)

% after rewrite_away_not
\bigwedge\left(
         \begin{array}{c}
         \bigvee\left(
                \begin{array}{c}
                4 \leq x_{51} \\
                x_{51}+2 \leq 1 \\
                3 \leq x_{52} \\
                x_{52}+2 \leq 2 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                4 \leq x_{51} \\
                x_{51}+2 \leq 3 \\
                4 \leq x_{52} \\
                x_{52}+2 \leq 3 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                2+\ell_{31} \leq 3 \\
                5+\ell_{32} \leq 7 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                2+\ell_{31} \leq x_{51} \\
                x_{51}+2 \leq 2 \\
                5+\ell_{32} \leq x_{52} \\
                x_{52}+2 \leq 5 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                6 \leq x_{51} \\
                x_{51}+2 \leq 3 \\
                8 \leq x_{52} \\
                x_{52}+2 \leq 7 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                4<x_{51} \\
                x_{51}+2<3 \\
                4<x_{52} \\
                x_{52}+2<3 \\
                \bigwedge\left(
                         \begin{array}{c}
                         4 \neq x_{51} \\
                         x_{51}+2 \neq 3 \\
                         4 \neq x_{52} \\
                         x_{52}+2 \neq 3 \\
                         \end{array}
                         \right) \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                6<x_{51} \\
                x_{51}+2<3 \\
                8<x_{52} \\
                x_{52}+2<7 \\
                \bigwedge\left(
                         \begin{array}{c}
                         6 \neq x_{51} \\
                         x_{51}+2 \neq 3 \\
                         8 \neq x_{52} \\
                         x_{52}+2 \neq 7 \\
                         \end{array}
                         \right) \\
                \end{array}
                \right) \\
         \end{array}
         \right)

% after rewrite_relations
\bigwedge\left(
         \begin{array}{c}
         \bigvee\left(
                \begin{array}{c}
                x_{51}-4 \geq 0 \\
                1-x_{51}+2 \geq 0 \\
                x_{52}-3 \geq 0 \\
                2-x_{52}+2 \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51}-4 \geq 0 \\
                3-x_{51}+2 \geq 0 \\
                x_{52}-4 \geq 0 \\
                3-x_{52}+2 \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                3-2+\ell_{31} \geq 0 \\
                7-5+\ell_{32} \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51}-2+\ell_{31} \geq 0 \\
                2-x_{51}+2 \geq 0 \\
                x_{52}-5+\ell_{32} \geq 0 \\
                5-x_{52}+2 \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51}-6 \geq 0 \\
                3-x_{51}+2 \geq 0 \\
                x_{52}-8 \geq 0 \\
                7-x_{52}+2 \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51}-4>0 \\
                3-x_{51}+2>0 \\
                x_{52}-4>0 \\
                3-x_{52}+2>0 \\
                \bigwedge\left(
                         \begin{array}{c}
                         \bigvee\left(
                                \begin{array}{c}
                                4-x_{51}>0 \\
                                x_{51}-4>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                x_{51}+2-3>0 \\
                                3-x_{51}+2>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                4-x_{52}>0 \\
                                x_{52}-4>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                x_{52}+2-3>0 \\
                                3-x_{52}+2>0 \\
                                \end{array}
                                \right) \\
                         \end{array}
                         \right) \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51}-6>0 \\
                3-x_{51}+2>0 \\
                x_{52}-8>0 \\
                7-x_{52}+2>0 \\
                \bigwedge\left(
                         \begin{array}{c}
                         \bigvee\left(
                                \begin{array}{c}
                                6-x_{51}>0 \\
                                x_{51}-6>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                x_{51}+2-3>0 \\
                                3-x_{51}+2>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                8-x_{52}>0 \\
                                x_{52}-8>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                x_{52}+2-7>0 \\
                                7-x_{52}+2>0 \\
                                \end{array}
                                \right) \\
                         \end{array}
                         \right) \\
                \end{array}
                \right) \\
         \end{array}
         \right)

% after rewrite_away_ops & min_max
\bigwedge\left(
         \begin{array}{c}
         \bigvee\left(
                \begin{array}{c}
                1 \times x_{51}+-4 \times 1 \geq 0 \\
                1 \times 1+-1 \times x_{51}+-2 \times 1 \geq 0 \\
                1 \times x_{52}+-3 \times 1 \geq 0 \\
                2 \times 1+-1 \times x_{52}+-2 \times 1 \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                1 \times x_{51}+-4 \times 1 \geq 0 \\
                3 \times 1+-1 \times x_{51}+-2 \times 1 \geq 0 \\
                1 \times x_{52}+-4 \times 1 \geq 0 \\
                3 \times 1+-1 \times x_{52}+-2 \times 1 \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                3 \times 1+-2 \times 1+-1 \times \ell_{31} \geq 0 \\
                7 \times 1+-5 \times 1+-1 \times \ell_{32} \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                1 \times x_{51}+-2 \times 1+-1 \times \ell_{31} \geq 0 \\
                2 \times 1+-1 \times x_{51}+-2 \times 1 \geq 0 \\
                1 \times x_{52}+-5 \times 1+-1 \times \ell_{32} \geq 0 \\
                5 \times 1+-1 \times x_{52}+-2 \times 1 \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                1 \times x_{51}+-6 \times 1 \geq 0 \\
                3 \times 1+-1 \times x_{51}+-2 \times 1 \geq 0 \\
                1 \times x_{52}+-8 \times 1 \geq 0 \\
                7 \times 1+-1 \times x_{52}+-2 \times 1 \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                1 \times x_{51}+-4 \times 1>0 \\
                3 \times 1+-1 \times x_{51}+-2 \times 1>0 \\
                1 \times x_{52}+-4 \times 1>0 \\
                3 \times 1+-1 \times x_{52}+-2 \times 1>0 \\
                \bigwedge\left(
                         \begin{array}{c}
                         \bigvee\left(
                                \begin{array}{c}
                                4 \times 1+-1 \times x_{51}>0 \\
                                1 \times x_{51}+-4 \times 1>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                1 \times x_{51}+2 \times 1+-3 \times 1>0 \\
                                3 \times 1+-1 \times x_{51}+-2 \times 1>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                4 \times 1+-1 \times x_{52}>0 \\
                                1 \times x_{52}+-4 \times 1>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                1 \times x_{52}+2 \times 1+-3 \times 1>0 \\
                                3 \times 1+-1 \times x_{52}+-2 \times 1>0 \\
                                \end{array}
                                \right) \\
                         \end{array}
                         \right) \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                1 \times x_{51}+-6 \times 1>0 \\
                3 \times 1+-1 \times x_{51}+-2 \times 1>0 \\
                1 \times x_{52}+-8 \times 1>0 \\
                7 \times 1+-1 \times x_{52}+-2 \times 1>0 \\
                \bigwedge\left(
                         \begin{array}{c}
                         \bigvee\left(
                                \begin{array}{c}
                                6 \times 1+-1 \times x_{51}>0 \\
                                1 \times x_{51}+-6 \times 1>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                1 \times x_{51}+2 \times 1+-3 \times 1>0 \\
                                3 \times 1+-1 \times x_{51}+-2 \times 1>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                8 \times 1+-1 \times x_{52}>0 \\
                                1 \times x_{52}+-8 \times 1>0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                1 \times x_{52}+2 \times 1+-7 \times 1>0 \\
                                7 \times 1+-1 \times x_{52}+-2 \times 1>0 \\
                                \end{array}
                                \right) \\
                         \end{array}
                         \right) \\
                \end{array}
                \right) \\
         \end{array}
         \right)

% after rewrite_linearize
\bigwedge\left(
         \begin{array}{c}
         \bigvee\left(
                \begin{array}{c}
                x_{51} \geq 4 \\
                -1 \times x_{51} \geq 1 \\
                x_{52} \geq 3 \\
                -1 \times x_{52} \geq 0 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51} \geq 4 \\
                -1 \times x_{51} \geq -1 \\
                x_{52} \geq 4 \\
                -1 \times x_{52} \geq -1 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                -1 \times \ell_{31} \geq -1 \\
                -1 \times \ell_{32} \geq -2 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                -1 \times \ell_{31}+x_{51} \geq 2 \\
                -1 \times x_{51} \geq 0 \\
                -1 \times \ell_{32}+x_{52} \geq 5 \\
                -1 \times x_{52} \geq -3 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51} \geq 6 \\
                -1 \times x_{51} \geq -1 \\
                x_{52} \geq 8 \\
                -1 \times x_{52} \geq -5 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51} \geq 5 \\
                -1 \times x_{51} \geq 0 \\
                x_{52} \geq 5 \\
                -1 \times x_{52} \geq 0 \\
                \bigwedge\left(
                         \begin{array}{c}
                         \bigvee\left(
                                \begin{array}{c}
                                -1 \times x_{51} \geq -3 \\
                                x_{51} \geq 5 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                x_{51} \geq 2 \\
                                -1 \times x_{51} \geq 0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                -1 \times x_{52} \geq -3 \\
                                x_{52} \geq 5 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                x_{52} \geq 2 \\
                                -1 \times x_{52} \geq 0 \\
                                \end{array}
                                \right) \\
                         \end{array}
                         \right) \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51} \geq 7 \\
                -1 \times x_{51} \geq 0 \\
                x_{52} \geq 9 \\
                -1 \times x_{52} \geq -4 \\
                \bigwedge\left(
                         \begin{array}{c}
                         \bigvee\left(
                                \begin{array}{c}
                                -1 \times x_{51} \geq -5 \\
                                x_{51} \geq 7 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                x_{51} \geq 2 \\
                                -1 \times x_{51} \geq 0 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                -1 \times x_{52} \geq -7 \\
                                x_{52} \geq 9 \\
                                \end{array}
                                \right) \\
                         \bigvee\left(
                                \begin{array}{c}
                                x_{52} \geq 6 \\
                                -1 \times x_{52} \geq -4 \\
                                \end{array}
                                \right) \\
                         \end{array}
                         \right) \\
                \end{array}
                \right) \\
         \end{array}
         \right)

% after rewrite_simplify
\bigwedge\left(
         \begin{array}{c}
         \bigvee\left(
                \begin{array}{c}
                x_{51} \geq 4 \\
                x_{52} \geq 3 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51} \geq 4 \\
                -1 \times x_{51} \geq -1 \\
                x_{52} \geq 4 \\
                -1 \times x_{52} \geq -1 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                -1 \times \ell_{31}+x_{51} \geq 2 \\
                -1 \times \ell_{32}+x_{52} \geq 5 \\
                -1 \times x_{52} \geq -3 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51} \geq 6 \\
                -1 \times x_{51} \geq -1 \\
                -1 \times x_{52} \geq -5 \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51} \geq 5 \\
                x_{52} \geq 5 \\
                \bigwedge\left(
                         \begin{array}{c}
                         \bigvee\left(
                                \begin{array}{c}
                                -1 \times x_{51} \geq -3 \\
                                x_{51} \geq 5 \\
                                \end{array}
                                \right) \\
                         x_{51} \geq 2 \\
                         \bigvee\left(
                                \begin{array}{c}
                                -1 \times x_{52} \geq -3 \\
                                x_{52} \geq 5 \\
                                \end{array}
                                \right) \\
                         x_{52} \geq 2 \\
                         \end{array}
                         \right) \\
                \end{array}
                \right) \\
         \bigvee\left(
                \begin{array}{c}
                x_{51} \geq 7 \\
                -1 \times x_{52} \geq -4 \\
                \bigwedge\left(
                         \begin{array}{c}
                         \bigvee\left(
                                \begin{array}{c}
                                -1 \times x_{51} \geq -5 \\
                                x_{51} \geq 7 \\
                                \end{array}
                                \right) \\
                         x_{51} \geq 2 \\
                         \bigvee\left(
                                \begin{array}{c}
                                x_{52} \geq 6 \\
                                -1 \times x_{52} \geq -4 \\
                                \end{array}
                                \right) \\
                         \end{array}
                         \right) \\
                \end{array}
                \right) \\
         \end{array}
         \right)

% New indexicals
% S3 not in {i | S3=i => L31 >= X51-1} /\ {i | S3=i => L32 >= X52-4} /\ (if X52=<3 then empty else Z^k}
o(3)-procedure(14,14,[int(-1)-1,
		      xmax(5,1)-2,
		      plus(1,2)-3,
		      sids(3,1,l(1),>=,3)-4,
		      intersection(4,13,[int(-4)-5,
					 xmax(5,2)-6,
					 plus(5,6)-7,
					 sids(3,1,l(2),>=,7)-8,
					 intersection(8,12,[int(3)-9,
							    xmin(5,2)-10,
							    minus(9,10)-11,
							    gez(11)-12])-13])-14],[o(5)]).

% X51 x X52 not in [inf..(1+L31),4..(4+L32)]
o(5)-procedure(9,9,[inf-1,
		    int(1)-2,
		    pmin(3,1,l(1))-3,
		    plus(2,3)-4,
		    int(4)-5,
		    pmin(3,1,l(2))-6,
		    plus(5,6)-7,
		    sup-8,
		    obox([1..4,5..7,1..8])-9],[o(3)]).

% MERGED GROUND INDEXICAL
o(5)-procedure(32,32,[inf-1,
		      int(6)-2,
		      int(5)-3,
		      sup-4,
		      obox([1..2,3..4,1..4])-5,
		      intersection(5,11,[obox([1..4,3..3,1..4])-6,
					 union(6,10,[int(1)-7,
						     obox([1..7,1..4,1..4])-8,
						     union(8,9,[obox([2..2,1..4,1..4])-9])-10])-11])-12,
		      union(12,31,[int(4)-13,
				   obox([1..13,1..13,1..4])-14,
				   intersection(14,22,[int(1)-15,
						       obox([1..15,1..4,1..4])-16,
						       union(16,17,[obox([13..13,1..4,1..4])-17])-18,
						       union(18,21,[obox([1..4,1..15,1..4])-19,
								    union(19,20,[obox([1..4,13..13,1..4])-20])-21])-22])-23,
				   union(23,30,[int(2)-24,
						obox([24..3,2..4,1..4])-25,
						union(25,29,[int(3)-26,
							     obox([1..26,1..24,1..4])-27,
							     union(27,28,[obox([24..26,24..26,1..4])-28])-29])-30])-31])-32],[]).


% old
o(5)-procedure(33,33,[inf-1,
		      int(6)-2,
		      int(5)-3,
		      sup-4,
		      obox([1..2,3..4,1..4])-5,
		      intersection(5,11,[obox([1..4,1..3,1..4])-6,
					 union(6,10,[obox([2..4,1..4,1..4])-7,
						     union(7,9,[int(1)-8,
								obox([1..8,1..4,1..4])-9])-10])-11])-12,
		      union(12,32,[int(4)-13,
				   obox([1..13,1..13,1..4])-14,
				   intersection(14,23,[obox([13..4,1..4,1..4])-15,
						       union(15,17,[int(1)-16,
								    obox([1..16,1..4,1..4])-17])-18,
						       union(18,22,[obox([1..4,13..4,1..4])-19,
								    union(19,21,[int(1)-20,
										 obox([1..4,1..20,1..4])-21])-22])-23])-24,
				   union(24,31,[int(2)-25,
						obox([25..3,2..4,1..4])-26,
						union(26,30,[int(3)-27,
							     obox([1..27,1..25,1..4])-28,
							     union(28,29,[obox([25..27,25..27,1..4])-29])-30])-31])-32])-33],[]).



% REGRESSION TEST

| ?- example(S,U,V).
% TOTAL #insns 121
% sweeping object 3
% found witness [2,5,3]
% found witness [2,5,4]
% sweeping object 5
% found witness [5,1,5]
% found witness [9,6,5]
% sweeping object 3
S in 3..4,
U in 5..9,
V in 1..6 ? 

/* now with more sharing */
% TOTAL #insns 97
% sweeping object 3
% found witness [2,5,3]
% found witness [2,5,4]
% sweeping object 5
% found witness [5,1,5]
% found witness [9,6,5]
% sweeping object 3
S in 3..4,
U in 5..9,
V in 1..6 ? 
