:- use_module(library(between)).
:- use_module(library(lists)).
:- use_module(library(clpfd)).
:- ensure_loaded(geosr).

% Each box is modelled as obj(L, W, H)
% Boxes are placed in containers of the same sizes
% The objective is to minimize the number of containers
% Data are checked in such away they fit in the containers
% Example : if the length of the item is greater than the container length
%           then take the container length as the maximum size
%           Same control for the width and the height

% Total volume = 7399713140!!

% main(geost, 50, 3).
% main(geosr, 50, 3).

main(Version, Max, Nbin) :-
        prolog_flag(informational, _, off),
        between(1, Max, N),
        call(Version, N/Nbin),
        fail.
main(_, _, _) :-
        prolog_flag(informational, _, on).

geosr(N/Nbin) :-
        Gabarit = [1200,800,1500],
        GabaritN = [1200,800,1500,Nbin],
        findall(SortKey-obj(L,W,H), clipped(L,W,H,SortKey,Gabarit,1), Objs1),
        keysort(Objs1, Objs2),
        prefix_length(Objs2, Objs2b, N),
        statistics(runtime,_),
        keys_and_values(Objs2b, _, Objs3),
        keys_and_values(Pairs1, Objs3, _), !,
        keysort(Pairs1, Pairs2),
        keyclumped(Pairs2, Groups),
        fix_shapes(Groups, 0),
        diffs_objects(Pairs1, Objects, GabaritN, 0),
        diffs_shapes(Groups, Shapes),
        numlist(1, N, OidList),
        geosr(Objects,
              Shapes,
              [(origin(O1,S1,D) ---> O1^x(D)+S1^t(D)),
               (end(O1,S1,D) ---> O1^x(D)+S1^t(D)+S1^l(D)),
               (overlap_sboxes(Dims, O1, S1, O2, S2) --->
                   forall(D, Dims,
                          end(O1,S1,D) #> origin(O2,S2,D) #/\
                          end(O2,S2,D) #> origin(O1,S1,D))),
               (oversize(O1,O2,S1,S2,D) ---> % amount by which two sboxes overlap in dim D
                   max(max(origin(O1,S1,D),origin(O2,S2,D)) - min(origin(O1,S1,D),origin(O2,S2,D)),
                       max(end(O1,S1,D),   end(O2,S2,D)) -    min(end(O1,S1,D),   end(O2,S2,D)))),
               (gravity(O1,Os) ---> % O1 is either on the floor on sitting on some other box
                   forall(S1,sboxes([O1^sid]),
                          (origin(O1,S1,4) #= 0 #\/
                           exists(O2,Os,
                                  O1^oid#\=O2^oid #/\
                                  exists(S2,sboxes([O2^sid]),
                                         overlap_sboxes([1,2,3],O1,S1,O2,S2) #/\
                                         origin(O1,S1,4) #= end(O2,S2,4)))))),
               forall(O1,objects(OidList),
                      gravity(O1,objects(OidList)) #/\
                      forall(S1,sboxes([O1^sid]),
                             forall(O2,objects(OidList),
                                    O1^oid #< O2^oid #=>
                                    forall(S2,sboxes([O2^sid]),
                                           overlap_sboxes([1,2,3],O1,S1,O2,S2) #=> 
                                           #\ overlap_sboxes([4],O1,S1,O2,S2) #/\ % essential! non-overlap explicitly!
                                           oversize(O1,O2,S1,S2,2) #=< 10 #/\
                                           oversize(O1,O2,S1,S2,3) #=< 10))))]),
        used_memory(Mem1),
        object_vars(Objects, Vars, []),
        labeling([], Vars),
        used_memory(Mem3),
        Mem2 is Mem3-Mem1,
        statistics(runtime,[_,Time]),
        fd_statistics(backtracks, Backtracks),
        portray_clause(Objects),
        format('N=~d, backtracks=~d, static memory=~d bytes, dynamic memory=~d bytes, time=~d msec.\n', [N,Backtracks,Mem1,Mem2,Time]), !.

geost(N/Nbin) :-
        Gabarit = [1200,800,1500],
        GabaritN = [1200,800,1500,Nbin],
        findall(SortKey-obj(L,W,H), clipped(L,W,H,SortKey,Gabarit,1), Objs1),
        keysort(Objs1, Objs2),
        prefix_length(Objs2, Objs2b, N),
        statistics(runtime,_),
        keys_and_values(Objs2b, _, Objs3),
        keys_and_values(Pairs1, Objs3, _), !,
        keysort(Pairs1, Pairs2),
        keyclumped(Pairs2, Groups),
        fix_shapes(Groups, 0),
        diffs_objects(Pairs1, Objects, GabaritN, 0),
        diffs_shapes(Groups, Shapes),
        ann_objects(Objects, Objects4, Shapes),

	gravity(Objects4, Objects4),
        overhang(Objects4, Objects4),
        geost(Objects, Shapes),

	used_memory(Mem1),
        object_vars(Objects, Vars, []),
        labeling([], Vars),
        used_memory(Mem3),
        Mem2 is Mem3-Mem1,
        statistics(runtime,[_,Time]),
        fd_statistics(backtracks, Backtracks),
        portray_clause(Objects),
        format('N=~d, backtracks=~d, static memory=~d bytes, dynamic memory=~d bytes, time=~d msec.\n', [N,Backtracks,Mem1,Mem2,Time]), !.

ann_objects([], [], _).
ann_objects([object(OID,SID,Xs)|Os3], [object(OID,SID,Xs,Ss)|Os4], Shapes) :-
        memberchk(sbox(SID,_,Ss), Shapes),
        ann_objects(Os3, Os4, Shapes).

gravity([], _).
gravity([O1|Os], Objects) :-
        O1 = object(_,_,Xs,_),
        Xs = [_,_,_,X4],
        X4 #= 0 #<=> X4Z,
        supports(Objects, O1, Support),
        sum([X4Z|Support], #>=, 1),
        gravity(Os, Objects).

supports([], _, []).
supports([O2|Os], O1, [B|Bs]) :-
        O1 = object(OID1,_,Xs1,Ss1),
        O2 = object(OID2,_,Xs2,Ss2),
        OID1 =\= OID2, !,
        Xs1 = [X11,X12,X13,X14],
        Xs2 = [X21,X22,X23,X24],
        Ss1 = [_,S12,S13,_],
        Ss2 = [_,S22,S23,S24],
        (   X11 #= X21 #/\
            X12+S12 #> X22 #/\
            X22+S22 #> X12 #/\
            X13+S13 #> X23 #/\
            X23+S23 #> X13 #/\
            X14 #= X24+S24 #<=> B
        ),
        supports(Os, O1, Bs).
supports([_|Os], O1, Bs) :-
        supports(Os, O1, Bs).

supports([], _, _, []).
supports([O2|Os], O1, Shapes, [B|Bs]) :-
        O1 = object(OID1,_,Xs1,Ss1),
        O2 = object(OID2,_,Xs2,Ss2),
        OID1 =\= OID2, !,
        Xs1 = [X11,X12,X13,X14],
        Xs2 = [X21,X22,X23,X24],
        Ss1 = [_,S12,S13,_],
        Ss2 = [_,S22,S23,S24],
        (   X11 #= X21 #/\
            X12+S12 #> X22 #/\
            X22+S22 #> X12 #/\
            X13+S13 #> X23 #/\
            X23+S23 #> X13 #/\
            X14 #= X24+S24 #<=> B
        ),
        supports(Os, O1, Shapes, Bs).
supports([_|Os], O1, Shapes, Bs) :-
        supports(Os, O1, Shapes, Bs).
        
overhang([], _).
overhang([O1|Os], Objects) :-
        overhang(Os, O1, Objects),
        overhang(Os, Objects).

overhang([], _, _).
overhang([O2|Os], O1, Objects) :-
        O1 = object(OID1,_,Xs1,Ss1),
        O2 = object(OID2,_,Xs2,Ss2),
        Xs1 = [X11,X12,X13,_],
        Xs2 = [X21,X22,X23,_],
        Ss1 = [_,S12,S13,_],
        Ss2 = [_,S22,S23,_],
        (   OID1 >= OID2 -> true
        ;   (   X11 #= X21 #/\
                X12+S12 #> X22 #/\
                X22+S22 #> X12 #/\
                X13+S13 #> X23 #/\
                X23+S23 #> X13 #=>
                max(max(X12,X22) - min(X12,X22),
                    max(X12+S12,   X22+S22) -    min(X12+S12,   X22+S22)) #=< 10 #/\
                max(max(X13,X23) - min(X13,X23),
                    max(X13+S13,   X23+S23) -    min(X13+S13,   X23+S23)) #=< 10
            )
        ),
        overhang(Os, O1, Objects).

used_memory(Mem) :-
        garbage_collect,
        statistics(global_stack, [G,_]),
        statistics(local_stack, [L,_]),
        statistics(trail, [T,_]),
        statistics(choice, [C,_]),
        Mem is G+L+T+C.

object_vars([]) --> [].
object_vars([object(_,_,[A,B,C,D])|L]) --> [A,B,C,D],
        object_vars(L).


fixall(NbUsed, Scale) :-
        % Model of containers
        % Objets are packed in containers
        % Each object is modelled with 4 decision variables {T, Y, X, Z}
        % T is the container and it ranges from 1 to the number of items
        % MaxLength = 1200,
        % MaxWidth = 800,
        % MaxHeight = 1500,
        Gabarit = [1200,800,1500],
        scale_gabarit(Gabarit, ScaledGabarit, Scale),
        findall(SortKey-obj(L,W,H), clipped(L,W,H,SortKey,ScaledGabarit,Scale), Objs1),
        keysort(Objs1, Objs2),
        keys_and_values(Objs2, _, Objs3),
        keys_and_values(Pairs1, Objs3, _),
        keysort(Pairs1, Pairs2),
        keyclumped(Pairs2, Groups),
        fix_shapes(Groups, 0),
        NbCont in 0..50,
        indomain(NbCont),
        fixall(Pairs1, Groups, Objects, ScaledGabarit, NbCont, NbUsed),
        report(Pairs1, Objects),
        true.

fixall(Pairs, Groups, Objects, Gabarit, NbCont, NbUsed) :-
        N0 is NbCont-1,
        append(Gabarit, [N0], GabaritN),
        diffs_objects(Pairs, Objects, GabaritN, 0),
        diffs_shapes(Groups, Shapes),
        % object_dvars(Objects, Ts, 1),
        % graph_colors(Ts, -1),
        geost(Objects, Shapes, [cumulative(true),longest_hole(true),parconflict(true),visavis(true),fixall(1,[object(_,min(1),[min(2),min(3),min(4),min(5)])])]),
        object_containers(Objects, Containers),
        max_member(Max, Containers),
        NbUsed is Max+1.

object_containers([], []).
object_containers([object(_,_,[X|_])|Os], [X|Xs]) :-
        object_containers(Os, Xs).
        

top(Pairs, Groups, Objects, Gabarit, N) :-
        fd_max(N, Nmax),
        N0 is Nmax-1,
        append(Gabarit, [N0], GabaritN),
        diffs_objects(Pairs, Objects, GabaritN, 0),
        diffs_shapes(Groups, Shapes),
        % object_dvars(Objects, Ts, 1),
        % graph_colors(Ts, -1),
        geost(Objects, Shapes, [fixall(1)]),
        search(Objects, 0, N).

search([], N, N) :- !.
search(Objs0, N0, N) :-
        place(Objs0, N0, Objs),
        N1 is N0+1,
        search(Objs, N1, N).

place([], _, []).
place([Obj|Objs0], T, Objs) :-
        Obj = object(_,_,[T,W,L,H]),
        fd_min(W, W),
        fd_min(L, L),
        fd_min(H, H), !,
        place(Objs0, T, Objs).
place([X|Objs0], T, [X|Objs]) :-
        place(Objs0, T, Objs).
        

graph_colors([], _).
graph_colors([X|Xs], Max0) :-
        X #=< Max0+1,
        Max #= max(Max0,X),
        graph_colors(Xs, Max).

diffs_objects([], [], _, _).
diffs_objects([obj(L,W,H)-SID|Pairs], [object(OID,SID,[T,Y,X,Z])|Objects], Gabarit, OID0) :-
        Gabarit = [Length,Width,Height,Cont],
        OID is OID0+1,
        Xmax is Length-L,
        Ymax is Width-W,
        Zmax is Height-H,
        Tmax is Cont-1,
        X in 0..Xmax,
        Y in 0..Ymax,
        Z in 0..Zmax,
        T in 0..Tmax,
        diffs_objects(Pairs, Objects, Gabarit, OID).

diffs_shapes([], []).
diffs_shapes([obj(L,W,H)-[SID|_]|Groups], [sbox(SID,[0,0,0,0],[1,W,L,H])|Shapes]) :-
        diffs_shapes(Groups, Shapes).

object_dvars([], [], [], [], _).
object_dvars([Obj-_|Pairs], [object(_,_,Origin)|Objects], [X|Xs], [L|Ls], D) :-
        arg(D, Obj, L),
        nth1(D, Origin, X),
        object_dvars(Pairs, Objects, Xs, Ls, D).
        
object_dvars([], [], _).
object_dvars([object(_,_,Origin)|Objects], [X|Xs], D) :-
        nth1(D, Origin, X),
        object_dvars(Objects, Xs, D).

fix_shapes([], _).
fix_shapes([_-Values|Groups], S0) :-
        S is S0+1,
        domain(Values, S, S),
        fix_shapes(Groups, S).

ensure_cps([], _, _).
ensure_cps([X|Xs], [N|Ns], W) :-
        ensure_cp(X, N, W),
        ensure_cps(Xs, Ns, W).

ensure_cp(X, Len, W) :-
        fd_min(X, Min),
        fd_max(X, Max),
        Mid is (Min+Max)>>1,
        (   Max-Min < Len*W -> true
        ;   X in Min..Mid,
            ensure_cp(X, Len, W)
        ;   Mid1 is Mid+1,
            X in Mid1..Max,
            ensure_cp(X, Len, W)
        ).

        
scale_gabarit([L1,W1,H1], [L2,W2,H2], S) :-
        L2 is L1//S,
        W2 is W1//S,
        H2 is H1//S.

clipped(L1, W1, H1, key(NVol,NW,NL,NH), [Lm,Wm,Hm], Scale) :-
        obj(L, W, H),
        L1 is min((L-1)//Scale+1,Lm),
        W1 is min((W-1)//Scale+1,Wm),
        H1 is min((H-1)//Scale+1,Hm),
        NW is -W1,
        NL is -L1,
        NH is -H1,
        NVol is NW*NL*NH.

report([], []).
report([Obj-_|Pairs], [object(_,_,[T,Y,X,Z])|Objects]) :-
        format('~q placed at [T,X,Y,Z]=~q\n', [Obj,[T,X,Y,Z]]),
        report(Pairs, Objects).


:- dynamic obj/3.
obj(300,155,130).
obj(300,155,130).
obj(300,155,130).
obj(300,155,130).
obj(275,190,100).
obj(275,190,100).
obj(275,190,100).
obj(210,150,155).
obj(210,150,155).
obj(245,110,110).
obj(270,145,110).
obj(270,145,110).
obj(270,145,110).
obj(377,260,157).
obj(380,260,165).
obj(720,397,50).
obj(275,260,150).
obj(275,260,150).
obj(275,260,150).
obj(275,260,150).
obj(275,260,150).
obj(860,45,50).
obj(860,45,50).
obj(860,45,50).
obj(860,45,50).
obj(860,45,50).
obj(860,45,50).
obj(860,45,50).
obj(860,45,50).
obj(860,45,50).
obj(860,45,50).
obj(583,547,50).
obj(670,170,165).
obj(300,170,100).
obj(300,170,100).
obj(300,170,100).
obj(300,170,100).
obj(300,170,100).
obj(300,170,100).
obj(300,170,100).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(580,340,340).
obj(377,260,157).
obj(377,260,157).
obj(370,85,60).
obj(370,85,60).
obj(370,85,60).
obj(370,85,60).
obj(370,85,60).
obj(370,85,60).
obj(380,260,165).
obj(380,260,165).
obj(583,547,50).
obj(583,547,50).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(580,340,340).
obj(580,340,340).
obj(580,340,340).
obj(580,340,340).
obj(580,340,340).
obj(580,340,340).
obj(580,340,340).
obj(580,340,340).
obj(377,260,157).
obj(377,260,157).
obj(520,595,170).
obj(520,170,200).
obj(520,170,200).
obj(520,170,200).
obj(520,170,200).
obj(520,170,200).
obj(920,200,200).
obj(920,200,200).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(580,340,340).
obj(580,340,340).
obj(580,340,340).
obj(377,260,157).
obj(520,595,170).
obj(520,595,170).
obj(380,260,165).
obj(720,1091,50).
obj(520,490,157).
obj(520,490,157).
obj(520,490,157).
obj(583,547,50).
obj(583,547,50).
obj(583,547,50).
obj(583,547,50).
obj(583,547,50).
obj(583,547,50).
obj(583,547,50).
obj(670,170,165).
obj(520,170,200).
obj(910,200,200).
obj(910,200,200).
obj(920,200,200).
obj(920,200,200).
obj(440,350,265).
obj(440,350,265).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(670,170,165).
obj(910,200,200).
obj(910,200,200).
obj(910,200,200).
obj(910,200,200).
obj(910,200,200).
obj(910,200,200).
obj(910,200,200).
obj(910,200,200).
obj(650,200,200).
obj(650,200,200).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(440,350,265).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1500,150,150).
obj(1505,155,50).
obj(1505,155,50).
obj(1505,155,50).
obj(1505,155,50).
obj(1505,155,50).
obj(1505,155,50).
obj(1505,155,50).
obj(1505,155,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(1000,200,200).
obj(500,425,155).
obj(500,425,155).
obj(500,425,155).
obj(500,425,155).
obj(500,425,155).
obj(500,425,155).
obj(500,425,155).
obj(500,425,155).
obj(500,425,155).
obj(500,425,155).
obj(720,20,20).
obj(720,20,20).
obj(1005,120,80).
obj(1005,155,45).
obj(1005,155,45).
obj(1005,155,45).
obj(1005,155,45).
obj(1005,155,45).
obj(1005,155,45).
obj(610,140,140).
obj(1200,90,90).
obj(90,85,50).
obj(90,85,50).
obj(90,85,50).
obj(90,85,50).
obj(90,85,50).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(490,100,100).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,25).
obj(2000,150,25).
obj(2000,150,25).
obj(2000,150,25).
obj(250,250,100).
obj(250,250,100).
obj(1500,150,25).
obj(1500,150,25).
obj(245,115,110).
obj(245,115,110).
obj(245,115,110).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(2000,150,50).
obj(610,140,140).
obj(610,140,140).
obj(480,420,125).
obj(670,395,130).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
obj(3000,40,40).
